from .binary import(
    fit_affine_scaling,
    warm_up_affine_scaling,
    fit_quadratic_scaling,
    warm_up_quadratic_scaling
)
from .vector_scaling import fit_vector_scaling, warm_up_vector_scaling
from .matrix_scaling import fit_matrix_scaling, warm_up_matrix_scaling



import math
import warnings
import numpy as np
from numba import njit
from typing import Callable, Tuple
from .utils import _fast_shuffle, _soft_threshold, _mcp_threshold, _mcp_penalty

L_FACTOR = 0.25
ETA_DIVISOR = 3.0

# Proximal operators
@njit(fastmath=True, cache=True)
def _prox_mcp(W: np.ndarray, k: int, step_size: float, reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float) -> None:
    """Inplace Minimax Concave Penalty (MCP) proximal operator for the weight matrix of our hierarchical Softmax regression.
    See "Nearly unbiased variable selection under minimax concave penalty" https://arxiv.org/abs/1002.4734
    """
    for j in range(k):
        W[j, k] = _mcp_threshold(W[j, k], reg_intercept, step_size) # intercepts
        W[j, j] = _mcp_threshold(W[j, j], reg_diagonal, step_size) # diagonal
        for l in range(j):
            W[j, l] = _mcp_threshold(W[j, l], reg_off_diagonal, step_size) # off-diagonal
            W[l, j] = _mcp_threshold(W[l, j], reg_off_diagonal, step_size) # off-diagonal

@njit(fastmath=True, cache=True)
def _prox_lasso(W: np.ndarray, k: int, step_size: float, reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float) -> None:
    """Inplace LASSO proximal operator for the weight matrix of our hierarchical Softmax regression."""
    thresh_intercept = step_size * reg_intercept
    thresh_diagonal = step_size * reg_diagonal
    thresh_off_diagonal = step_size * reg_off_diagonal

    for j in range(k):
        W[j, k] = _soft_threshold(W[j, k], thresh_intercept) # intercept
        W[j, j] = _soft_threshold(W[j, j], thresh_diagonal) # diagonal
        for l in range(j):
            W[j, l] = _soft_threshold(W[j, l], thresh_off_diagonal) # off-diagonal
            W[l, j] = _soft_threshold(W[l, j], thresh_off_diagonal) # off-diagonal

@njit(fastmath=True, cache=True)
def _prox_group_lasso(W: np.ndarray, k: int, step_size: float, reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float) -> None:
    """Inplace group-lasso proximal operator for the weight matrix of our hierarchical Softmax regression."""
    # Computing norms
    norm_intercept_sq = 0.0
    norm_diagonal_sq = 0.0
    norm_off_diagonal_sq = 0.0

    for j in range(k):
        norm_intercept_sq += W[j, k] ** 2
        norm_diagonal_sq += W[j, j] ** 2
        for l in range(j):
            norm_off_diagonal_sq += W[j, l] ** 2 + W[l, j] ** 2

    # Proximal step intercept
    if norm_intercept_sq > 0.0:
        shrinkage = max(0.0, 1.0 - step_size * reg_intercept / math.sqrt(norm_intercept_sq))
        for j in range(k):
            W[j, k] *= shrinkage

    # Proximal step diagonal
    if norm_diagonal_sq > 0.0:
        shrinkage = max(0.0, 1.0 - step_size * reg_diagonal / math.sqrt(norm_diagonal_sq))
        for j in range(k):
            W[j, j] *= shrinkage

    # Proximal step off-diagonal
    if norm_off_diagonal_sq > 0.0:
        shrinkage = max(0.0, 1.0 - step_size * reg_off_diagonal / math.sqrt(norm_off_diagonal_sq))
        for j in range(k):
            for l in range(j):
                W[j, l] *= shrinkage
                W[l, j] *= shrinkage

@njit(fastmath=True, cache=True)
def _prox_ridge(W: np.ndarray, k: int, step_size: float, reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float) -> None:
    """Inplace ridge proximal operator for the weight matrix of our hierarchical Softmax regression."""
    factor_intercept = 1.0 / (1.0 + 2.0 * step_size * reg_intercept)
    factor_diagonal = 1.0 / (1.0 + 2.0 * step_size * reg_diagonal)
    factor_off_diagonal = 1.0 / (1.0 + 2.0 * step_size * reg_off_diagonal)

    for j in range(k):
        W[j, k] *= factor_intercept # intercept
        W[j, j] *= factor_diagonal # diagonal
        for l in range(j):
            W[j, l] *= factor_off_diagonal # off-diagonal
            W[l, j] *= factor_off_diagonal # off-diagonal

@njit(fastmath=True, cache=True)
def _prox_none(W: np.ndarray, k: int, step_size: float, reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float) -> None:
    pass

# Logit computation
@njit(fastmath=True, cache=True)
def _compute_logits(
    xi: np.ndarray, a: float, W: np.ndarray, k: int, k_plus_1: int,
    logits: np.ndarray, exp_logits: np.ndarray
) -> float:
    """
    Computes numerically stable exp(logits) for a single sample.

    This function modifies logits and exp_logits in-place.

    Returns:
        The sum of the exponentiated logits (sum_exp).
    """
    # Part 1: Compute logits and find the maximum
    max_logit = -np.inf
    for j in range(k):
        logit = a * xi[j]
        for l in range(k_plus_1):
            logit += W[j, l] * xi[l]
        logits[j] = logit
        if logit > max_logit:
            max_logit = logit

    # Part 2: Compute stable exp(logits) and their sum
    sum_exp = 0.0
    for j in range(k):
        exp_logit = math.exp(logits[j] - max_logit)
        exp_logits[j] = exp_logit
        sum_exp += exp_logit
        
    return sum_exp

# Stopping functions
@njit(fastmath=True, cache=True)
def _compute_primal_dual_ridge(
    X: np.ndarray, y: np.ndarray, a: float, W: np.ndarray,
    reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float
) -> Tuple[float, float]:
    """Computes primal and (approximate) dual objectives for our structured matrix scaling convex problem with ridge regularization.
    The dual is only approximate since global scaling parameter a is not regularized. This sometimes leads to dual_obj > primal_obj.
    """
    n, k_plus_1 = X.shape
    k = k_plus_1 - 1
    
    primal_loss = 0.0
    dual_loss = 0.0
    s_W = np.zeros((k, k_plus_1))
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, W, k, k_plus_1, logits, exp_logits
        )

        for j in range(k):
            prob = exp_logits[j] / sum_exp
            alpha_ij = -prob
            dual_loss -= prob * math.log(prob)
            if j == yi:
                alpha_ij += 1.0
                primal_loss -= math.log(prob)
            for l in range(k_plus_1):
                s_W[j, l] += alpha_ij * xi[l]

    primal_reg = 0.0
    for j in range(k):
        primal_reg += reg_intercept * W[j, k]**2
        primal_reg += reg_diagonal * W[j, j]**2
        for l in range(j):
            primal_reg += reg_off_diagonal * W[j, l]**2
            primal_reg += reg_off_diagonal * W[l, j]**2

    primal_obj = (primal_loss / n) + primal_reg

    dual_reg = 0.0
    n_squared = float(n * n)
    factor_intercept = 1.0 / (4.0 * reg_intercept * n_squared)
    factor_diagonal = 1.0 / (4.0 * reg_diagonal * n_squared)
    factor_off_diagonal = 1.0 / (4.0 * reg_off_diagonal * n_squared)
    for j in range(k):
        dual_reg -= factor_intercept * s_W[j, k]**2
        dual_reg -= factor_diagonal * s_W[j, j]**2
        for l in range(j):
            dual_reg -= factor_off_diagonal * s_W[j, l]**2
            dual_reg -= factor_off_diagonal * s_W[l, j]**2

    dual_obj = (dual_loss / n) + dual_reg
    
    return primal_obj, dual_obj

@njit(fastmath=True, cache=True)
def _compute_primal_dual_group_lasso(
    X: np.ndarray, y: np.ndarray, a: float, W: np.ndarray,
    reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float
) -> Tuple[float, float]:
    """Computes primal and (approximate) dual objectives for our structured matrix scaling convex problem with group-LASSO regularization.
    The dual is only approximate since global scaling parameter a is not regularized. This sometimes leads to dual_obj > primal_obj.
    """
    n, k_plus_1 = X.shape
    k = k_plus_1 - 1
    
    primal_loss = 0.0
    dual_loss = 0.0
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, W, k, k_plus_1, logits, exp_logits
        )
        
        for j in range(k):
            prob = exp_logits[j] / sum_exp
            dual_loss -= prob * math.log(prob)
            if j == yi:
                primal_loss -= math.log(prob)

    norm_intercept_sq = 0.0
    norm_diagonal_sq = 0.0
    norm_off_diagonal_sq = 0.0
    for j in range(k):
        norm_intercept_sq += W[j, k] ** 2
        norm_diagonal_sq += W[j, j] ** 2
        for l in range(j):
            norm_off_diagonal_sq += W[j, l] ** 2 + W[l, j] ** 2

    primal_reg = 0.0
    primal_reg += reg_intercept * math.sqrt(norm_intercept_sq)
    primal_reg += reg_diagonal * math.sqrt(norm_diagonal_sq)
    primal_reg += reg_off_diagonal * math.sqrt(norm_off_diagonal_sq)

    primal_obj = (primal_loss / n) + primal_reg

    dual_reg = 0.0
    dual_obj = (dual_loss / n) + dual_reg

    return primal_obj, dual_obj

@njit(fastmath=True, cache=True)
def _compute_primal_dual_lasso(
    X: np.ndarray, y: np.ndarray, a: float, W: np.ndarray,
    reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float
) -> Tuple[float, float]:
    """Computes primal and (approximate) dual objectives for our structured matrix scaling convex problem with LASSO regularization.
    The dual is only approximate since global scaling parameter a is not regularized. This sometimes leads to dual_obj > primal_obj.
    """
    n, k_plus_1 = X.shape
    k = k_plus_1 - 1
    
    primal_loss = 0.0
    dual_loss = 0.0
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, W, k, k_plus_1, logits, exp_logits
        )
        
        for j in range(k):
            prob = exp_logits[j] / sum_exp
            dual_loss -= prob * math.log(prob)
            if j == yi:
                primal_loss -= math.log(prob)

    primal_reg = 0.0
    for j in range(k):
        primal_reg += reg_intercept * np.abs(W[j, k])
        primal_reg += reg_diagonal * np.abs(W[j, j])
        for l in range(j):
            primal_reg += reg_off_diagonal * np.abs(W[j, l])
            primal_reg += reg_off_diagonal * np.abs(W[l, j])

    primal_obj = (primal_loss / n) + primal_reg

    dual_reg = 0.0
    dual_obj = (dual_loss / n) + dual_reg

    return primal_obj, dual_obj

@njit(fastmath=True, cache=True)
def _compute_primal_dual_mcp(
    X: np.ndarray, y: np.ndarray, a: float, W: np.ndarray,
    reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float
) -> Tuple[float, float]:
    """Computes primal objective for our structured matrix scaling convex problem with MCP regularization.
    With MCP penalty, the dual cannot be evaluated. Only primal is used for stopping.
    """
    n, k_plus_1 = X.shape
    k = k_plus_1 - 1
    
    primal_loss = 0.0

    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, W, k, k_plus_1, logits, exp_logits
        )
        
        primal_loss -= math.log(exp_logits[yi] / sum_exp)

    primal_reg = 0.0
    for j in range(k):
        primal_reg += _mcp_penalty(W[j, k], reg_intercept)
        primal_reg += _mcp_penalty(W[j, j], reg_diagonal)
        for l in range(j):
            primal_reg += _mcp_penalty(W[j, l], reg_off_diagonal)
            primal_reg += _mcp_penalty(W[l, j], reg_off_diagonal)

    primal_obj = (primal_loss / n) + primal_reg

    return primal_obj, -np.inf

@njit(fastmath=True, cache=True)
def _compute_primal_dual_none(
    X: np.ndarray, y: np.ndarray, a: float, W: np.ndarray,
    reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float
) -> Tuple[float, float]:
    """Computes primal objective for our structured matrix scaling convex problem without regularization.
    Since no regularization is applied, the dual cannot be evaluated. Only primal is used for stopping.
    """
    n, k_plus_1 = X.shape
    k = k_plus_1 - 1
    
    primal_loss = 0.0

    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, W, k, k_plus_1, logits, exp_logits
        )
        
        primal_loss -= math.log(exp_logits[yi] / sum_exp)

    return primal_loss, -np.inf

# SAGA solver
@njit(fastmath=True, cache=True)
def _saga_matrix_scaling(
    X: np.ndarray, y: np.ndarray, init_scaling: float, reg_intercept: float, reg_diagonal: float, reg_off_diagonal: float,
    max_iter: int, tol: float, prox_op: Callable[[np.ndarray, int, float, float, float, float], None],  stop_op: Callable[..., float]
) -> Tuple[float, np.ndarray, int]:
    """Solves our structured matrix scaling convex optimization problem using SAGA.

    Args:
        X (np.ndarray): un-calibrated logits.
        y (np.ndarray): un-calibrated labels (not one-hot encoded).
        init_scaling (float): initial value for global scaling parameter.
        reg_intercept (float): regularization strenght applied to intercept parameters.
        reg_diagonal (float): regularization strenght applied to diagonal parameters.
        reg_off_diagonal (float): regularization strenght applied to off-diagonal parameters.
        max_iter (int): max number of dataset passes before SAGA is stopped.
        tol (float): tolerance used for stopping.
            Stopping is based on relative difference between primal and (approximate) dual objectives when available.
            Otherwise relative difference between consecutive primal objectives is used.
        prox_op (Callable[[np.ndarray, int, float, float], None]): proximal operator used to regularize.
        stop_op (Callable[..., float]): function that computes the primal and dual objectives for stopping.

    Returns:
        Tuple[float, np.ndarray, int]: global scaling parameter a, weight matrix W, number of epochs performed.
    """
    n, k_plus_1 = X.shape
    k = k_plus_1 - 1

    a = init_scaling
    W = np.zeros((k, k_plus_1))

    previous_objective = np.inf
    samples_seen = 1

    gradient_weights_memory = np.zeros((n, k))
    avg_gradient = np.zeros((k, k_plus_1))
    avg_gradient_a = 0.0
    permutation = np.arange(n, dtype=np.int32)
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    L = L_FACTOR * np.max(np.sum(X * X, axis=1))
    step_size = 1.0 / (ETA_DIVISOR * L)

    for epoch in range(1, max_iter + 1):
        _fast_shuffle(permutation)
        
        for i in permutation:
            xi, yi = X[i], y[i]

            sum_exp = _compute_logits(
                xi, a, W, k, k_plus_1, logits, exp_logits
            )

            update_a = 0.0
            for j in range(k):
                prob = exp_logits[j] / sum_exp
                gradient_weight = prob - (1.0 if j == yi else 0.0)
                
                delta_grad = gradient_weight - gradient_weights_memory[i, j]
                gradient_weights_memory[i, j] = gradient_weight
                
                update_a += delta_grad * xi[j]

                for l in range(k_plus_1):
                    update = delta_grad * xi[l]
                    W[j, l] -= step_size * (update + avg_gradient[j, l])
                    if epoch == 1:
                        avg_gradient[j, l] += (update - avg_gradient[j, l]) / samples_seen
                    else:
                        avg_gradient[j, l] += update / n

            a -= step_size * (update_a + avg_gradient_a)
            if epoch == 1:
                avg_gradient_a += (update_a - avg_gradient_a) / samples_seen
                samples_seen += 1
            else:
                avg_gradient_a += update_a / n

            prox_op(W, k, step_size, reg_intercept, reg_diagonal, reg_off_diagonal)

        primal_obj, dual_obj = stop_op(X, y, a, W, reg_intercept, reg_diagonal, reg_off_diagonal)
        if dual_obj == -np.inf:
            gap = abs(primal_obj - previous_objective)
            if gap <= tol * max(1.0, abs(primal_obj)):
                return a, W, epoch
            previous_objective = primal_obj
        else:
            # Global scaling parameter a is not regularized, the duality gap is not exact and can be negative.
            # For robustness, we track the absolute gap.
            gap = abs(primal_obj - dual_obj)
            if gap <= tol * max(1.0, abs(primal_obj)):
                return a, W, epoch

    return a, W, max_iter

# Dictionnary mappings
MATRIX_PROX_MAP = {
    "mcp": _prox_mcp,
    "lasso": _prox_lasso,
    "group_lasso": _prox_group_lasso,
    "ridge": _prox_ridge,
    None: _prox_none,
}

MATRIX_STOP_MAP = {
    "mcp": _compute_primal_dual_mcp,
    "lasso": _compute_primal_dual_lasso,
    "group_lasso": _compute_primal_dual_group_lasso,
    "ridge": _compute_primal_dual_ridge,
    None: _compute_primal_dual_none,
}

# Public API functions
def fit_matrix_scaling(
    X: np.ndarray, y: np.ndarray, init_scaling: float, penalty: str, reg_intercept: float,
    reg_diagonal: float, reg_off_diagonal: float, max_iter: int, tol: float
) -> np.ndarray:
    """Public API to fit matrix scaling using SAGA."""
    k = X.shape[1] - 1
    try:
        prox_op = MATRIX_PROX_MAP[penalty]
        stop_op = MATRIX_STOP_MAP[penalty]
    except KeyError:
        raise ValueError(f"Unknown penalty: {penalty}")

    a, W, last_iter = _saga_matrix_scaling(
        X, y, init_scaling, reg_intercept, reg_diagonal, reg_off_diagonal,
        max_iter, tol, prox_op, stop_op
    )
    if last_iter == max_iter:
        warnings.warn("SAGA reached max_iter without convergence", RuntimeWarning)

    return W + np.hstack([np.eye(k) * a, np.zeros((k, 1))])

def warm_up_matrix_scaling():
    """Public API to warm up matrix scaling."""
    X = np.random.randn(10, 3).astype(np.float32)
    y = np.random.randint(0, 3, size=10)
    for penalty in [None, 'mcp', 'lasso', 'ridge']:
        _saga_matrix_scaling(X, y, 1.0/1.5, 3.0/10.0, 3.0/10.0, 2.0*3.0/10.0, 10, 1e-5, MATRIX_PROX_MAP[penalty], MATRIX_STOP_MAP[penalty])



import warnings
from typing import Callable, Optional, List

import torch
import sklearn
import logging
import numpy as np
from probmetrics.saga import (
    fit_affine_scaling,
    warm_up_affine_scaling,
    fit_quadratic_scaling,
    warm_up_quadratic_scaling,
    fit_vector_scaling,
    warm_up_vector_scaling,
    fit_matrix_scaling,
    warm_up_matrix_scaling
)
from probmetrics.utils import binary_probs_to_logits, multiclass_probs_to_logits, validate_probabilities
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.calibration import CalibratedClassifierCV
from sklearn.model_selection import StratifiedKFold, GridSearchCV
from torch import nn

from probmetrics.distributions import CategoricalDistribution, CategoricalProbs, CategoricalLogits

logger = logging.getLogger(__name__)

class Calibrator(BaseEstimator, ClassifierMixin):
    """
    Calibrator base class. To implement,
    - override at least one of _fit_impl and _fit_torch_impl
    - override at least one of predict_proba and predict_proba_torch
    """

    def __init__(self):
        assert self.__class__.fit == Calibrator.fit
        assert self.__class__.fit_torch == Calibrator.fit_torch

    def fit(self, X, y) -> 'Calibrator':
        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        self.classes_ = list(range(X.shape[-1]))

        if self.__class__._fit_impl != Calibrator._fit_impl:
            self._fit_impl(X, y)
            return self

        if self.__class__._fit_torch_impl != Calibrator._fit_torch_impl:
            self._fit_torch_impl(y_pred=CategoricalProbs(torch.as_tensor(X)),
                                 y_true_labels=torch.as_tensor(y, dtype=torch.long))
            return self

        raise NotImplementedError()

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        raise NotImplementedError()

    def fit_torch(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor) -> 'Calibrator':
        assert isinstance(y_true_labels, torch.Tensor)
        assert isinstance(y_pred, CategoricalDistribution)
        # default implementation, using sklearn
        self.classes_ = list(range(y_pred.get_n_classes()))

        if self.__class__._fit_torch_impl != Calibrator._fit_torch_impl:
            self._fit_torch_impl(y_pred, y_true_labels)
            return self

        if self.__class__._fit_impl != Calibrator._fit_impl:
            self._fit_impl(y_pred.get_probs().detach().cpu().numpy(), y_true_labels.detach().cpu().numpy())
            return self

        raise NotImplementedError()

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        raise NotImplementedError()

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        if self.__class__.predict_proba_torch != Calibrator.predict_proba_torch:
            return self.predict_proba_torch(CategoricalProbs(torch.as_tensor(X))).get_probs().numpy()

        raise NotImplementedError()

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        if self.__class__.predict_proba != Calibrator.predict_proba:
            y_pred_probs = y_pred.get_probs()
            probs = self.predict_proba(y_pred_probs.detach().cpu().numpy())
            return CategoricalProbs(torch.as_tensor(probs, device=y_pred_probs.device, dtype=y_pred_probs.dtype))

        raise NotImplementedError()

    def predict(self, X):
        y_probs = self.predict_proba(X)
        class_idxs = np.argmax(y_probs, axis=-1)
        return np.asarray(self.classes_)[class_idxs]


class ApplyToLogitsCalibrator(Calibrator):
    def __init__(self, calib: BaseEstimator):
        super().__init__()
        self.calib = calib

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        self.calib_ = sklearn.base.clone(self.calib)
        self.calib_.fit(np.log(X + 1e-10), y)

    def predict_proba(self, X):
        return self.calib_.predict_proba(np.log(X + 1e-10))


def bisection_search(f: Callable[[float], float], a: float, b: float, n_steps: int):
    for _ in range(n_steps):
        c = a + 0.5 * (b - a)
        f_c = f(c)
        if f_c > 0:
            b = c
        else:
            a = c

    return 0.5 * (a + b)


class TemperatureScalingCalibrator(Calibrator):
    def __init__(self, opt: str = 'bisection', max_bisection_steps: int = 30, lr: float = 0.1, max_iter: int = 200,
                 use_inv_temp: bool = True, inv_temp_init: float = 1 / 1.5):
        super().__init__()
        self.lr = lr
        self.max_bisection_steps = max_bisection_steps
        self.max_iter = max_iter
        self.use_inv_temp = use_inv_temp
        self.inv_temp_init = inv_temp_init
        self.opt = opt

    def _get_loss_grad(self, invtemp: float, logits: torch.Tensor, y: torch.Tensor):
        part_1 = torch.mean(torch.sum(logits * torch.softmax(invtemp * logits, dim=-1), dim=-1))
        part_2 = torch.mean(logits[torch.arange(logits.shape[0]), y])
        return (part_1 - part_2).item()

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        logits = y_pred.get_logits()
        labels = y_true_labels

        if self.opt in ['lbfgs', 'lbfgs_line_search']:
            self._fit_lbfgs(logits, labels)
        elif self.opt == 'bisection':
            self._fit_bisection(logits, labels)
        else:
            raise ValueError(f'Unknown optimizer "{self.opt}"')

        # print(f'{self.invtemp_=}')

    def _fit_lbfgs(self, logits: torch.Tensor, labels: torch.Tensor):
        # following https://github.com/gpleiss/temperature_scaling/blob/master/temperature_scaling.py
        param = nn.Parameter(
            torch.ones(1, device=logits.device) * (self.inv_temp_init if self.use_inv_temp else 1 / self.inv_temp_init))
        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.LBFGS([param], lr=self.lr, max_iter=self.max_iter,
                                      line_search_fn='strong_wolfe' if self.opt == 'lbfgs_line_search' else None)

        def eval():
            optimizer.zero_grad()
            y_pred = logits * param[:, None] if self.use_inv_temp else logits / param[:, None]
            loss = criterion(y_pred, labels)
            loss.backward()
            return loss

        optimizer.step(eval)

        self.invtemp_ = param.item() if self.use_inv_temp else 1 / param.item()

    def _fit_bisection(self, logits: torch.Tensor, labels: torch.Tensor):
        objective_grad = lambda u, l=logits, tar=labels: self._get_loss_grad(np.exp(u), l, tar)

        # should reach about float32 accuracy
        # need log_2(32) = 5 steps to get to length 1 and then 24 more steps to get to float32 epsilon (2^{-24})
        self.invtemp_ = np.exp(bisection_search(objective_grad, a=-16, b=16, n_steps=self.max_bisection_steps))
        # print(f'{self.invtemp_=}')

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        return CategoricalLogits(self.invtemp_ * y_pred.get_logits())


### New binary calibrators ###

class BinaryLogisticCalibrator(Calibrator):
    """
    Binary post-hoc calibration with linear, affine or quadratic scaling 
    on the binary logits using sklearn's LogisticRegression.

    This class fits a model of the form:
    P(y=1) = sigma(calibrated_logit)
    
    Where `calibrated_logit` depends on the `type`:
    - 'linear':    calibrated_logit = w * logit (Temperature scaling with inverse temperature)
    - 'affine':    calibrated_logit = b + w * logit (Platt scaling)
    - 'quadratic': calibrated_logit = b + w1 * logit + w2 * logit^2
    """

    VALID_TYPES = ['linear', 'affine', 'quadratic']

    def __init__(self, type: str = 'affine'):
        """
        Args:
            type (str, optional): The type of scaling. 
                One of ['linear', 'affine', 'quadratic']. 
                Defaults to 'affine'.
        """
        super().__init__()
        
        if type not in self.VALID_TYPES:
            raise ValueError(
                f"Invalid type '{type}'. Must be one of {self.VALID_TYPES}"
            )
        self.type = type
        self.cal_ = None

    def _get_logits(self, X: np.ndarray) -> np.ndarray:
        logits = binary_probs_to_logits(X)
        logits = logits.reshape(-1,1)
        if self.type == 'affine':
            # Adding a constant term to logit vectors.
            logits = np.hstack([np.ones((len(X), 1)), logits])
        elif self.type == 'quadratic':
            # Adding a constant and quadratic term to logit vectors.
            logits = np.hstack([np.ones((len(X), 1)), logits, np.square(logits)])
        return logits

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        """Fits the logistic regression calibrator."""
        from sklearn.linear_model import LogisticRegression
        self.cal_ = LogisticRegression(penalty=None, fit_intercept=False)

        features = self._get_logits(X)
        self.cal_.fit(features, y)

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        """Generates calibrated probability estimates."""
        if self.cal_ is None:
            raise RuntimeError("Calibrator has not been fitted yet. Call fit() before predict_proba().")

        features = self._get_logits(X)
        return self.cal_.predict_proba(features)

class AffineScalingCalibrator(Calibrator):
    """
    Binary post-hoc calibration with Platt scaling (~binary logistic regression) $sigmoid(ax+b)$ on the logits x.
    Uses the SAGA algorithm to fit the scaling and intercept parameters a & b.
    Numba functions are warmed-up the first time the class is initialized (~1s).
    A penalty (one of ["mcp", "lasso", "ridge"]) can be applied to the intercept parameter b, with regularization strength lambda_intercept.
    To apply no regularization, leave penalty to None.
    """
    _warmed_up = False

    def __init__(
        self,
        penalty: str = None,
        lambda_intercept: float = 0.0,
        max_iter: int = 200,
        tol: float = 1e-4,
        print_init_info: bool = True
    ):
        super().__init__()
        self.penalty = penalty
        self.lambda_intercept = lambda_intercept
        self.max_iter = max_iter
        self.tol = tol
        self.print_init_info = print_init_info

        self.w = None

        if not AffineScalingCalibrator._warmed_up:
            if self.print_init_info:
                print("First AffineScalingCalibrator instantiation - warming up Numba functions...")
            warm_up_affine_scaling()
            AffineScalingCalibrator._warmed_up = True
            if self.print_init_info:
                print("Warmed up!")

    def _get_logits(self, X: np.ndarray) -> np.ndarray: 
        logits = binary_probs_to_logits(X)
        logits = logits.reshape(-1,1)
        # Adding a constant term to logit vectors to fit the intercept b.
        logits = np.hstack([np.ones((len(X), 1)), logits])
        return logits

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        logits = self._get_logits(X)
        self.w = fit_affine_scaling(
            logits, y,
            penalty=self.penalty,
            reg_intercept=self.lambda_intercept,
            max_iter=self.max_iter,
            tol=self.tol,
        )

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        if self.w is None:
            raise RuntimeError("Calibrator has not been fitted yet. Call fit() before predict_proba().")

        logits = self._get_logits(X)
        logits = logits.dot(self.w).reshape(-1,1)
        probs = 1.0 / (1.0 + np.exp(-logits))
        return np.hstack([1-probs, probs])


class QuadraticScalingCalibrator(Calibrator):
    """
    Binary post-hoc calibration with quadratic scaling $sigmoid(a + bx + cx^2)$ on the logits x.
    Uses the SAGA algorithm to fit the intercept, scaling and quadratic parameters a, b & c.
    Numba functions are warmed-up the first time the class is initialized (~1s).
    A penalty (one of ["mcp", "lasso", "ridge"]) can be applied to the intercept and quadratic parameters a & c,
    with respective regularization strength lambda_intercept and lambda_quadratic.
    To apply no regularization, leave penalty to None.
    """
    _warmed_up = False

    def __init__(
        self,
        penalty: str = None,
        lambda_intercept: float = 0.0,
        lambda_quadratic: float = 0.0,
        max_iter: int = 200,
        tol: float = 1e-4,
        print_init_info: bool = True
    ):
        super().__init__()
        self.penalty = penalty
        self.lambda_intercept = lambda_intercept
        self.lambda_quadratic = lambda_quadratic
        self.max_iter = max_iter
        self.tol = tol
        self.print_init_info = print_init_info

        self.w = None

        if not QuadraticScalingCalibrator._warmed_up:
            if self.print_init_info:
                print("First QuadraticScalingCalibrator instantiation - warming up Numba functions...")
            warm_up_quadratic_scaling()
            QuadraticScalingCalibrator._warmed_up = True
            if self.print_init_info:
                print("Warmed up!")

    def _get_logits(self, X: np.ndarray) -> np.ndarray: 
        logits = binary_probs_to_logits(X)
        logits = logits.reshape(-1,1)
        # Adding a constant and quadratic term to logit vectors.
        logits = np.hstack([np.ones((len(X), 1)), logits, np.square(logits)])
        return logits

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:        
        logits = self._get_logits(X)
        self.w = fit_quadratic_scaling(
            logits, y,
            penalty=self.penalty,
            reg_intercept=self.lambda_intercept,
            reg_quadratic=self.lambda_quadratic,
            max_iter=self.max_iter,
            tol=self.tol,
        )

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        if self.w is None:
            raise RuntimeError("Calibrator has not been fitted yet. Call fit() before predict_proba().")

        logits = self._get_logits(X)
        logits = logits.dot(self.w).reshape(-1,1)
        probs = 1.0 / (1.0 + np.exp(-logits))
        return np.hstack([1-probs, probs])

##################################


### New multiclass calibrators ###

class SVSCalibrator(Calibrator):
    """
    Multiclass post-hoc calibration with a structured scaling scheme $softmax((aI + diag(v))x + b)$ on the logits x.

    Numba functions are warmed-up the first time the class is initialized with the 'saga' optimizer.
        
    A penalty (one of [None, "mcp", "lasso", "ridge"]) is applied to the intercept (b) and vector scaling (v) parameters,
    with respective regularization strength lambda_intercept*(k**rho)/(n**tau) and lambda_diagonal*(k**rho)/(n**tau).
    Note that the general scaling parameter (a) is not regularized, just like in classical temperature scaling.

    [!] solver = 'bfgs' only supports 'ridge' penalty.
    [!] For solver = 'bfgs', max_iter and tol are ignored and default torchmin parameters are used.
    """
    _warmed_up_saga = False
    _ALLOWED_PENALTIES = [None, "mcp", "lasso", "ridge"]
    _ALLOWED_OPTIMIZERS = ["saga", "bfgs"]

    def __init__(
        self,
        penalty: str = 'ridge',
        rho: float = 1.0,
        tau: float = 1.0,
        lambda_intercept: float = 1.0,
        lambda_diagonal: float = 1.0,
        max_iter: int = 500,
        tol: float = 1e-5,
        opt: str = 'saga',
        ts_preprocessing: bool = True,
        print_init_info: bool = True
    ):
        super().__init__()

        if opt not in self._ALLOWED_OPTIMIZERS:
            raise ValueError(f"Optimizer '{opt}' not recognized. Choose from {self._ALLOWED_OPTIMIZERS}.")
        if penalty not in self._ALLOWED_PENALTIES:
            raise ValueError(f"Penalty '{penalty}' not recognized. Choose from {self._ALLOWED_PENALTIES}.")

        if opt == 'saga' and not SVSCalibrator._warmed_up_saga:
            if print_init_info:
                logger.info("First SVSCalibrator instantiation with 'saga' - warming up Numba functions...")
            warm_up_vector_scaling()
            SVSCalibrator._warmed_up_saga = True
            if print_init_info:
                logger.info("Warmed up!")

        if opt == 'bfgs' and penalty != 'ridge':
            raise ValueError("The 'bfgs' optimizer only supports the 'ridge' penalty.")

        if opt == 'bfgs':
            try:
                import torchmin
            except ImportError:
                raise ImportError("The 'torchmin' package is required for the 'bfgs' optimizer. Please install it.")

        self.penalty = penalty
        self.rho = rho
        self.tau = tau
        self.lambda_intercept = lambda_intercept
        self.lambda_diagonal = lambda_diagonal
        self.max_iter = max_iter
        self.tol = tol
        self.opt = opt
        self.ts_preprocessing = ts_preprocessing
        self.print_init_info = print_init_info

        self.a = None
        self.v = None
        self.b = None
        self.ts = None

    def _get_logits(self, X: np.ndarray) -> tuple[np.ndarray, int, int]:
        n, k = X.shape
        logits = multiclass_probs_to_logits(X)
        return logits, n, k

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        validate_probabilities(X)

        if self.ts_preprocessing:
            self.ts = get_calibrator('ts-mix')
            self.ts.fit(X, y)
            probs_to_calibrate = self.ts.predict_proba(X)
        else:
            self.ts = None
            probs_to_calibrate = X
        
        logits, n, k = self._get_logits(probs_to_calibrate)

        self.a = 1.0
        self.v = np.zeros(k)
        self.b = np.zeros(k)

        reg_intercept = self.lambda_intercept * (k**self.rho) / (n**self.tau)
        reg_diagonal = self.lambda_diagonal * (k**self.rho) / (n**self.tau)

        if self.opt == 'saga':
            self.a, self.v, self.b = fit_vector_scaling(
                logits, y,
                init_scaling=1.0,
                penalty=self.penalty,
                reg_intercept=reg_intercept,
                reg_diagonal=reg_diagonal,
                max_iter=self.max_iter,
                tol=self.tol,
            )

        elif self.opt == 'bfgs':
            self._fit_bfgs(
                torch.tensor(logits), torch.tensor(y),
                num_classes=k,
                reg_intercept=reg_intercept,
                reg_diagonal=reg_diagonal
            )

    def _fit_bfgs(self, logits: torch.Tensor, y: torch.Tensor, num_classes, reg_intercept, reg_diagonal):
        import torchmin

        a = torch.tensor([1.0], device=logits.device, requires_grad=True)
        v = torch.zeros(num_classes, device=logits.device, requires_grad=True)
        b = torch.zeros(num_classes, device=logits.device, requires_grad=True)

        params = torch.cat([a, v, b])

        def loss(params):
            a, v, b = params.split([1, num_classes, num_classes])
            logloss = torch.nn.functional.cross_entropy((a + v) * logits + b, y)
            reg = reg_intercept * torch.dot(b, b) + reg_diagonal * torch.dot(v, v)
            return logloss + reg

        res = torchmin.bfgs._minimize_bfgs(loss, params)

        a, v, b = res.x.split([1, num_classes, num_classes])

        self.a = a.item()
        self.v = v.detach().cpu().numpy()
        self.b = b.detach().cpu().numpy()

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        validate_probabilities(X)

        if self.a is None:
            raise RuntimeError("Calibrator has not been fitted yet. Call fit() before predict_proba().")

        if self.ts_preprocessing:
            probs_to_calibrate = self.ts.predict_proba(X) 
        else:
            probs_to_calibrate = X
        
        logits, _, _ = self._get_logits(probs_to_calibrate)
        
        calibrated_logits = (self.a + self.v.reshape(1,-1)) * logits + self.b.reshape(1,-1)
        calibrated_logits -= np.max(calibrated_logits, axis=1, keepdims=True)
        exp_logits = np.exp(calibrated_logits)
        return exp_logits / np.sum(exp_logits, axis=1, keepdims=True)


class SMSCalibrator(Calibrator):
    """
    Multiclass post-hoc calibration with a structured scaling scheme $softmax( (aI + diag(v) + off_diag(M) )x + b)$ on the logits x.

    Numba functions are warmed-up the first time the class is initialized with the 'saga' optimizer.

    A penalty (one of ["mcp", "lasso", "group_lasso", "ridge"]) is applied to the intercept (b), vector scaling (v) and off diagonal (M) parameters,
    with respective regularization strength lambda_intercept*(k**rho)/(n**tau), lambda_diagonal*(k**rho)/(n**tau) and lambda_off_diagonal*((k*(k-1))**rho)/(n**tau).
    Note that the general scaling parameter (a) is not regularized, just like in classical temperature scaling.

    [!] The 'bfgs' solver only supports 'ridge' penalty.
    [!] If 'bfgs' solver is used, max_iter and tol are ignored and default torchmin parameters are used.
    """
    _warmed_up_saga = False
    _ALLOWED_PENALTIES = [None, "mcp", "lasso", "group_lasso", "ridge"]
    _ALLOWED_OPTIMIZERS = ["saga", "bfgs"]

    def __init__(
        self,
        penalty: str = 'ridge',
        rho: float = 1.0,
        tau: float = 1.0,
        lambda_intercept: float = 1.0,
        lambda_diagonal: float = 1.0,
        lambda_off_diagonal: float = 1.0,
        max_iter: int = 500,
        tol: float = 1e-5,
        opt: str = 'saga',
        ts_preprocessing: bool = True,
        print_init_info: bool = True
    ):
        super().__init__()

        if opt not in self._ALLOWED_OPTIMIZERS:
            raise ValueError(f"Optimizer '{opt}' not recognized. Choose from {self._ALLOWED_OPTIMIZERS}.")
        if penalty not in self._ALLOWED_PENALTIES:
            raise ValueError(f"Penalty '{penalty}' not recognized. Choose from {self._ALLOWED_PENALTIES}.")

        if opt == 'saga' and not SMSCalibrator._warmed_up_saga:
            if print_init_info:
                logger.info("First SMSCalibrator instantiation with 'saga' - warming up Numba functions...")
            warm_up_matrix_scaling()
            SMSCalibrator._warmed_up_saga = True
            if print_init_info:
                logger.info("Warmed up!")

        if opt == 'bfgs' and penalty != 'ridge':
            raise ValueError("The 'bfgs' optimizer only supports the 'ridge' penalty.")

        if opt == 'bfgs':
            try:
                import torchmin
            except ImportError:
                raise ImportError("The 'torchmin' package is required for the 'bfgs' optimizer. Please install it.")

        self.penalty = penalty
        self.rho = rho
        self.tau = tau
        self.lambda_intercept = lambda_intercept
        self.lambda_diagonal = lambda_diagonal
        self.lambda_off_diagonal = lambda_off_diagonal
        self.max_iter = max_iter
        self.tol = tol
        self.opt = opt
        self.ts_preprocessing = ts_preprocessing
        self.print_init_info = print_init_info

        self.W = None
        self.ts = None

    def _get_logits(self, X: np.ndarray) -> tuple[np.ndarray, int, int]:
        n, k = X.shape
        logits = multiclass_probs_to_logits(X)
        # Adding a constant term to logit vectors to fit the intercept.
        logits = np.hstack([logits, np.ones((n, 1))])
        return logits, n, k

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        """Fits the calibration matrix W."""
        validate_probabilities(X)

        if self.ts_preprocessing:
            self.ts = get_calibrator('ts-mix')
            self.ts.fit(X, y)
            probs_to_calibrate = self.ts.predict_proba(X)
        else:
            self.ts = None
            probs_to_calibrate = X

        logits, n, k = self._get_logits(probs_to_calibrate)

        self.W = np.hstack([np.eye(k), np.zeros((k, 1))])

        reg_intercept=self.lambda_intercept * (k**self.rho) / (n**self.tau)
        reg_diagonal=self.lambda_diagonal * (k**self.rho) / (n**self.tau)
        reg_off_diagonal=self.lambda_off_diagonal * ((k*(k-1))**self.rho) / (n**self.tau)

        if self.opt == 'saga':
            self.W = fit_matrix_scaling(
                logits, y,
                penalty=self.penalty, max_iter=self.max_iter, tol=self.tol, init_scaling=1.0,
                reg_intercept=reg_intercept, reg_diagonal=reg_diagonal, reg_off_diagonal=reg_off_diagonal
            )

        elif self.opt == 'bfgs':
            self._fit_bfgs(
                torch.tensor(logits), torch.tensor(y), num_classes=k,
                reg_intercept=reg_intercept, reg_diagonal=reg_diagonal, reg_off_diagonal=reg_off_diagonal
            )

    def _fit_bfgs(self, logits: torch.Tensor, y: torch.Tensor, num_classes, reg_intercept, reg_diagonal, reg_off_diagonal):
        import torchmin

        a = torch.tensor([1.0], device=logits.device, dtype=logits.dtype)
        W = torch.zeros((num_classes, num_classes + 1), device=logits.device, dtype=logits.dtype)

        diag_mask = torch.hstack([torch.eye(num_classes), torch.zeros((num_classes, 1))])
        off_diag_mask = torch.hstack([torch.ones((num_classes, num_classes)) - torch.eye(num_classes), torch.zeros((num_classes, 1))])
        intercept_mask = torch.hstack([torch.zeros((num_classes, num_classes)), torch.ones((num_classes, 1))])

        def loss(params):
            a, W = params.split([1, num_classes*(num_classes+1)])
            a = a.squeeze()
            W = W.view(num_classes, num_classes+1)
            loss = torch.nn.functional.cross_entropy(logits @ (a*diag_mask + W).T, y)
            reg = reg_intercept*torch.square(intercept_mask * W).sum()
            reg += reg_diagonal*torch.square(diag_mask * W).sum()
            reg += reg_off_diagonal*torch.square(off_diag_mask * W).sum()
            return loss + reg

        params = torch.cat([a, W.view(-1)])
        method = "l-bfgs" if params.numel() > 1000 else "bfgs"
        res = torchmin.minimize(loss, params, method=method)

        a, W = res.x.split([1, num_classes*(num_classes+1)])
        W = W.view(num_classes, num_classes+1)

        self.W = a.item()*diag_mask.numpy() + W.detach().cpu().numpy()

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        validate_probabilities(X)

        if self.W is None:
            raise RuntimeError("Calibrator has not been fitted yet. Call fit() before predict_proba().")

        if self.ts_preprocessing:
            probs_to_calibrate = self.ts.predict_proba(X) 
        else:
            probs_to_calibrate = X
        
        logits, _, _ = self._get_logits(probs_to_calibrate)
        calibrated_logits = logits @ self.W.T
        calibrated_logits -= np.max(calibrated_logits, axis=1, keepdims=True)
        exp_logits = np.exp(calibrated_logits)
        return exp_logits / np.sum(exp_logits, axis=1, keepdims=True)

##################################


### Default logistic calibrator, compatible with binary and multiclass ###

class LogisticCalibrator(Calibrator):
    """
    By default, uses 'affine-scaling' for binary classification and 'sms' for multiclass classification.
    """

    VALID_BINARY_TYPES = ['linear', 'affine', 'quadratic']
    VALID_MULTICLASS_TYPES = ['svs', 'sms']

    def __init__(self, binary_type: str = 'affine', multiclass_type: str = 'sms'):
        super().__init__()
        
        if binary_type not in self.VALID_BINARY_TYPES:
            raise ValueError(
                f"Invalid binary type '{binary_type}'. Must be one of {self.VALID_BINARY_TYPES}"
            )
        self.binary_type = binary_type

        if multiclass_type not in self.VALID_MULTICLASS_TYPES:
            raise ValueError(
                f"Invalid multiclass type '{multiclass_type}'. Must be one of {self.VALID_MULTICLASS_TYPES}"
            )
        self.multiclass_type = multiclass_type

        self.cal_ = None

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        n_classes = len(np.unique(y))
        
        if n_classes == 2:
            if self.binary_type == 'affine':
                self.cal_ = BinaryLogisticCalibrator(type='affine')
            elif self.binary_type == 'linear':
                self.cal_ = BinaryLogisticCalibrator(type='linear')
            elif self.binary_type == 'quadratic':
                self.cal_ = BinaryLogisticCalibrator(type='quadratic')

        else:
            if self.multiclass_type == 'sms':
                self.cal_ = SMSCalibrator(opt='bfgs')
            elif self.multiclass_type == 'svs':
                self.cal_ = SVSCalibrator(opt='bfgs')

        self.cal_.fit(X, y)
        return self

    def predict(self, X):
        if self.cal_ is None:
            raise RuntimeError("LogisticCalibrator must be fitted before prediction.")
        return self.cal_.predict(X)

    def predict_proba(self, X):
        if self.cal_ is None:
            raise RuntimeError("LogisticCalibrator must be fitted before prediction.")
        return self.cal_.predict_proba(X)

##################################


class GuoTemperatureScalingCalibrator(Calibrator):
    # adapted from https://github.com/gpleiss/temperature_scaling/blob/master/temperature_scaling.py
    def __init__(self):
        super().__init__()
        self.temperature = nn.Parameter(torch.ones(1) * 1.5)

    def temperature_scale(self, logits):
        """
        Perform temperature scaling on logits
        """
        # Expand temperature to match the size of logits
        temperature = self.temperature.unsqueeze(1).expand(logits.size(0), logits.size(1))
        return logits / temperature

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        logits = y_pred.get_logits()
        labels = y_true_labels

        nll_criterion = nn.CrossEntropyLoss().cuda()

        # Next: optimize the temperature w.r.t. NLL
        optimizer = torch.optim.LBFGS([self.temperature], lr=0.01, max_iter=50)

        def eval():
            optimizer.zero_grad()
            loss = nll_criterion(self.temperature_scale(logits), labels)
            loss.backward()
            return loss

        optimizer.step(eval)

        return self

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        with torch.no_grad():
            return CategoricalLogits(y_pred.get_logits() / self.temperature)


class AutoGluonTemperatureScalingCalibrator(Calibrator):
    # adapted from
    # https://github.com/autogluon/autogluon/blob/c1181326cf6b7e3b27a7420273f1a82808d939e2/core/src/autogluon/core/calibrate/temperature_scaling.py#L9
    # https://github.com/autogluon/autogluon/blob/28a242ebe8d55ba770c991b9db153ab4623c9abd/tabular/src/autogluon/tabular/trainer/abstract_trainer.py#L4433-L4457
    def __init__(self, init_val: float = 1, max_iter: int = 200, lr: float = 0.1):
        super().__init__()
        self.init_val = init_val
        self.max_iter = max_iter
        self.lr = lr
        self.temperature = init_val

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        y_val_tensor = y_true_labels
        temperature_param = torch.nn.Parameter(torch.ones(1).fill_(self.init_val))
        logits = y_pred.get_logits()

        is_invalid = torch.isinf(logits).any().tolist()
        if is_invalid:
            return

        nll_criterion = torch.nn.CrossEntropyLoss()
        optimizer = torch.optim.LBFGS([temperature_param], lr=self.lr, max_iter=self.max_iter)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.99)

        optimizer_trajectory = []

        if self.init_val != 1.0:
            # need to check 1.0 as well since AutoGluon does it outside
            optimizer_trajectory.append((nll_criterion(logits, y_val_tensor).item(), 1.0))

        def temperature_scale_step():
            optimizer.zero_grad()
            temp = temperature_param.unsqueeze(1).expand(logits.size(0), logits.size(1))
            new_logits = logits / temp
            loss = nll_criterion(new_logits, y_val_tensor)
            loss.backward()
            scheduler.step()
            optimizer_trajectory.append((loss.item(), temperature_param.item()))
            return loss

        optimizer.step(temperature_scale_step)

        try:
            best_loss_index = np.nanargmin(np.array(optimizer_trajectory)[:, 0])
        except ValueError:
            self.temperature = 1.0
            return
        temperature_scale = float(np.array(optimizer_trajectory)[best_loss_index, 1])

        if np.isnan(temperature_scale) or temperature_scale <= 0.0:
            self.temperature = 1.0
            return

        self.temperature = temperature_scale

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        with torch.no_grad():
            return CategoricalLogits(y_pred.get_logits() / self.temperature)


class AutoGluonInverseTemperatureScalingCalibrator(Calibrator):
    # adapted from
    # https://github.com/autogluon/autogluon/blob/c1181326cf6b7e3b27a7420273f1a82808d939e2/core/src/autogluon/core/calibrate/temperature_scaling.py#L9
    # but optimizing the inverse temperature instead
    def __init__(self, init_val: float = 1, max_iter: int = 200, lr: float = 0.1):
        super().__init__()
        self.init_val = init_val
        self.max_iter = max_iter
        self.lr = lr
        self.inv_temp = init_val

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        y_val_tensor = y_true_labels
        inv_temperature_param = torch.nn.Parameter(torch.ones(1).fill_(self.init_val))
        logits = y_pred.get_logits()

        is_invalid = torch.isinf(logits).any().tolist()
        if is_invalid:
            return

        nll_criterion = torch.nn.CrossEntropyLoss()
        optimizer = torch.optim.LBFGS([inv_temperature_param], lr=self.lr, max_iter=self.max_iter)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.99)

        optimizer_trajectory = []

        def temperature_scale_step():
            optimizer.zero_grad()
            inv_temp = inv_temperature_param.unsqueeze(1).expand(logits.size(0), logits.size(1))
            new_logits = logits * inv_temp
            loss = nll_criterion(new_logits, y_val_tensor)
            loss.backward()
            scheduler.step()
            optimizer_trajectory.append((loss.item(), inv_temperature_param.item()))
            return loss

        optimizer.step(temperature_scale_step)

        try:
            best_loss_index = np.nanargmin(np.array(optimizer_trajectory)[:, 0])
        except ValueError:
            return
        inv_temperature_scale = float(np.array(optimizer_trajectory)[best_loss_index, 1])

        if np.isnan(inv_temperature_scale):
            return

        self.inv_temp = inv_temperature_scale

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        with torch.no_grad():
            return CategoricalLogits(y_pred.get_logits() * self.inv_temp)


class TorchUncertaintyTemperatureScalingCalibrator(TemperatureScalingCalibrator):
    # adapted from
    # https://github.com/ENSTA-U2IS-AI/torch-uncertainty/blob/3a021d2e34e183b8aad3a0345e6d750c08c72af3/torch_uncertainty/post_processing/calibration/scaler.py#L1
    # https://github.com/ENSTA-U2IS-AI/torch-uncertainty/blob/3a021d2e34e183b8aad3a0345e6d750c08c72af3/torch_uncertainty/post_processing/calibration/temperature_scaler.py#L9
    def __init__(self, init_val: float = 1, lr: float = 0.1, max_iter: int = 100):
        super().__init__(opt='lbfgs', lr=lr, max_iter=max_iter, use_inv_temp=False, inv_temp_init=1. / init_val)
        # need to save these values here to comply with sklearn cloneability conventions
        self.init_val = init_val
        self.lr = lr
        self.max_iter = max_iter


class NetcalTemperatureScalingCalibrator(Calibrator):
    # this one does nothing due to https://github.com/EFS-OpenSource/calibration-framework/issues/61
    def __init__(self):
        super().__init__()

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        from netcal.scaling import TemperatureScaling

        self.cal_ = TemperatureScaling()
        if X.shape[1] == 2:
            # binary, convert
            X = X[:, 1]
        self.cal_.fit(X, y, random_state=0)

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        if X.shape[1] == 2:
            # binary, convert
            X = X[:, 1]
        result = self.cal_.transform(X)
        if len(result.shape) == 1:
            return np.stack([1. - result, result], axis=1)
        elif result.shape[1] == 1:
            return np.concatenate([1. - result, result], axis=1)

        return result


class TorchcalTemperatureScalingCalibrator(Calibrator):
    # adapted from
    # https://github.com/rishabh-ranjan/torchcal/blob/3fb65f6423d33d680cd68c7f40a0259d41e8fb0b/torchcal.py#L8
    def __init__(self):
        super().__init__()
        self.temperature = 1.0

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        import torchmin
        y_pred_logits = y_pred.get_logits()

        temp = torch.ones(1, device=y_pred_logits.device)

        def loss(t):
            return torch.nn.functional.cross_entropy(y_pred_logits / t, y_true_labels)

        res = torchmin.minimize(loss, temp, method='newton-exact')
        if not res.success:
            warnings.warn(
                f"{self.__class__}: {res.message} Not updating calibrator params."
            )
        else:
            temp = res.x

        self.temperature = temp.item()

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        return CategoricalLogits(y_pred.get_logits() / self.temperature)


class TorchcalVectorScalingCalibrator(Calibrator):
    # adapted from
    # https://github.com/rishabh-ranjan/torchcal/blob/3fb65f6423d33d680cd68c7f40a0259d41e8fb0b/torchcal.py#L8
    def __init__(self):
        super().__init__()

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        import torchmin
        y_pred_logits = y_pred.get_logits()

        num_classes = y_pred_logits.shape[1]

        self.temp = torch.ones(num_classes, device=y_pred_logits.device)
        self.bias = torch.zeros(num_classes, device=y_pred_logits.device)

        def loss(temp_bias):
            temp, bias = temp_bias.split([num_classes, num_classes])
            return torch.nn.functional.cross_entropy(y_pred_logits / temp + bias, y_true_labels)
        temp_bias = torch.cat([self.temp, self.bias])
        res = torchmin.minimize(loss, temp_bias, method="bfgs")
        if not res.success:
            warnings.warn(
                f"{self.__class__}: {res.message} Not updating calibrator params."
            )
        else:
            self.temp, self.bias = res.x.split([num_classes, num_classes])

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        return CategoricalLogits(y_pred.get_logits() / self.temp + self.bias)


class TorchcalMatrixScalingCalibrator(Calibrator):
    def __init__(self):
        super().__init__()

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        import torchmin
        y_pred_logits = y_pred.get_logits()

        num_classes = y_pred_logits.shape[1]

        num_params = num_classes*(num_classes + 1)
        if num_params <= 1000:
            method = "bfgs"
        else:
            method = "l-bfgs"

        self.itemp = torch.eye(num_classes, num_classes, device=y_pred_logits.device, dtype=y_pred_logits.dtype)
        self.bias = torch.zeros(num_classes, device=y_pred_logits.device, dtype=y_pred_logits.dtype)

        def loss(itemp_bias):
            itemp, bias = itemp_bias.split([num_classes**2, num_classes])
            itemp = itemp.view(num_classes, num_classes)
            return torch.nn.functional.cross_entropy(y_pred_logits @ itemp + bias, y_true_labels)

        itemp_bias = torch.cat([self.itemp.view(-1), self.bias])
        res = torchmin.minimize(loss, itemp_bias, method=method)
        if not res.success:
            warnings.warn(
                f"{self.__class__}: {res.message} Not updating calibrator params."
            )
        else:
            self.itemp, self.bias = res.x.split(
                [num_classes**2, num_classes]
            )
            self.itemp = self.itemp.view(num_classes, num_classes)

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        return CategoricalLogits(y_pred.get_logits() @ self.itemp + self.bias)


class CenteredIsotonicRegressionCalibrator(Calibrator):
    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        assert not np.any(np.isnan(X))
        # print(f'{np.unique(y)=}')
        # print(f'{np.unique(X)=}')
        # we have to use float64 since with float32 it can happen that somewhere internally
        # a rounding error occurs (?) and interp1d thinks that it is asked to extrapolate at the boundary value,
        # which causes it to output nan values,
        # which are seen as a separate possible value but (y==value) is empty because nan==nan is false,
        # so an error occurs because the method attempts to take an average of an empty set
        X = X.astype(np.float64)
        y = y.astype(np.float64)
        from cir_model import CenteredIsotonicRegression
        self.cal_ = CenteredIsotonicRegression()
        self.cal_.fit(X[:, 1], y)
        self.min_ = np.min(X[:, 1])
        self.max_ = np.max(X[:, 1])

    def predict_proba(self, X):
        # have to clip since CenteredIsotonicRegression refuses to extrapolate (?)
        X = X.astype(np.float64)
        X = np.clip(X, self.min_, self.max_)
        pred_probs = self.cal_.transform(X[:, 1])
        pred_probs = np.clip(pred_probs, 0.0, 1.0)
        return np.stack([1. - pred_probs, pred_probs], axis=-1)


class BinaryVennAbersCalibrator(Calibrator):
    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        from venn_abers import VennAbers
        self.va_ = VennAbers()
        self.va_.fit(X, y)

    def predict_proba(self, X):
        return self.va_.predict_proba(X)[0]


class VennAbersCalibrator(Calibrator):
    def __init__(self, use_ovo: bool = False):
        super().__init__()
        self.use_ovo = use_ovo

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        self.X_ = np.copy(X)
        self.y_ = np.copy(y)

    def predict_proba(self, X):
        from venn_abers import VennAbersCalibrator
        va = VennAbersCalibrator()
        return va.predict_proba(p_cal=self.X_, y_cal=self.y_, p_test=X, p0_p1_output=False,
                                va_type='one_vs_one' if self.use_ovo else 'one_vs_all')


class MulticlassOneVsOneCalibrator(Calibrator):
    def __init__(self, binary_calibrator: BaseEstimator):
        super().__init__()
        self.binary_calibrator = binary_calibrator

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        self.n_classes_ = X.shape[-1]
        self.bin_calibs_ = []

        if self.n_classes_ == 2:
            # binary classification
            bin_calib = sklearn.base.clone(self.binary_calibrator)
            bin_calib.fit(X, y)
            self.bin_calibs_.append(bin_calib)
        else:
            for i in range(self.n_classes_):
                for j in range(i + 1, self.n_classes_):
                    idxs = np.logical_or((y == i), (y == j))
                    # idxs = np.arange(X.shape[0])
                    bin_probs = np.stack([X[idxs, j], X[idxs, i]], axis=-1)
                    bin_probs += 1e-30
                    bin_probs /= np.sum(bin_probs, axis=-1, keepdims=True)
                    # print(f'{np.any(np.isnan(bin_probs))=}')
                    bin_labels = (y[idxs] == i).astype(np.int32)
                    bin_calib = sklearn.base.clone(self.binary_calibrator)
                    bin_calib.fit(bin_probs, bin_labels)
                    self.bin_calibs_.append(bin_calib)

    def predict_proba(self, X):
        if self.n_classes_ == 2:
            # binary classification
            return self.bin_calibs_[0].predict_proba(X)
        else:
            # use PKPD formula in http://proceedings.mlr.press/v60/manokhin17a/manokhin17a.pdf
            pair_probs = [[None] * self.n_classes_ for i in range(self.n_classes_)]
            multi_probs = []
            calib_idx = 0
            for i in range(self.n_classes_):
                for j in range(i + 1, self.n_classes_):
                    bin_probs = np.stack([X[:, j], X[:, i]], axis=-1)
                    bin_probs += 1e-30
                    bin_probs /= np.sum(bin_probs, axis=-1, keepdims=True)
                    pred_probs = self.bin_calibs_[calib_idx].predict_proba(bin_probs)
                    pair_probs[i][j] = pred_probs[:, 1]
                    pair_probs[j][i] = pred_probs[:, 0]
                    # if i == 0 and j == 1:
                    #     print(f'{bin_probs=}')
                    #     print(f'{pred_probs=}')
                    calib_idx += 1

            for i in range(self.n_classes_):
                sum_inv_probs = sum([1. / (1e-30 + pair_probs[i][j]) for j in range(self.n_classes_) if j != i])
                multi_probs.append(1. / np.clip(sum_inv_probs - (self.n_classes_ - 2), 1e-30, np.inf))

            multi_probs = np.stack(multi_probs, axis=-1)
            multi_probs = np.clip(multi_probs, a_min=1e-30, a_max=np.inf)
            multi_probs = multi_probs / np.sum(multi_probs, axis=-1, keepdims=True)
            return multi_probs


class MulticlassOneVsRestCalibrator(Calibrator):
    def __init__(self, binary_calibrator: BaseEstimator):
        super().__init__()
        self.binary_calibrator = binary_calibrator

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        self.n_classes_ = X.shape[-1]
        self.bin_calibs_ = []

        X[np.isnan(X)] = 0.0

        if self.n_classes_ == 2:
            # binary classification
            bin_calib = sklearn.base.clone(self.binary_calibrator)
            bin_calib.fit(X, y)
            self.bin_calibs_.append(bin_calib)
        else:
            for i in range(self.n_classes_):
                pos_probs = X[:, i]
                neg_probs = 1. - pos_probs
                bin_probs = np.stack([neg_probs, pos_probs], axis=-1)
                bin_labels = (y == i).astype(np.int32)
                bin_calib = sklearn.base.clone(self.binary_calibrator)
                bin_calib.fit(bin_probs, bin_labels)
                self.bin_calibs_.append(bin_calib)

    def predict_proba(self, X):
        X[np.isnan(X)] = 0.0
        if self.n_classes_ == 2:
            # binary classification
            return self.bin_calibs_[0].predict_proba(X)
        else:
            multi_probs = []
            for i in range(self.n_classes_):
                pos_probs = X[:, i]
                neg_probs = 1. - pos_probs
                bin_probs = np.stack([neg_probs, pos_probs], axis=-1)
                pred_probs = self.bin_calibs_[i].predict_proba(bin_probs)
                multi_probs.append(pred_probs[:, 1])
            multi_probs = np.stack(multi_probs, axis=-1)
            multi_probs = np.clip(multi_probs, a_min=1e-30, a_max=np.inf)
            multi_probs = multi_probs / np.sum(multi_probs, axis=-1, keepdims=True)
            return multi_probs


class IdentityEstimator(ClassifierMixin, BaseEstimator):
    def __init__(self, n_classes: int):
        self.n_classes = n_classes
        # somehow it can fail if we don't do the np.asarray()
        self.classes_ = np.asarray(list(range(n_classes)))

    def fit(self, X, y):
        # self.classes_ = np.unique(y)  # if we do this we have a problem for missing classes
        return self

    def predict_proba(self, X):
        return X

    def predict(self, X):
        # having this as a dummy here for the FrozenEstimator solution
        # which hovewer doesn't work right now
        raise NotImplementedError()


class PreprocessingCalibratorWrapper(Calibrator):
    def __init__(self, calibrator: Calibrator, pre_calib: Optional[Calibrator] = None, l2_normalize: bool = False):
        super().__init__()
        self.calibrator = calibrator
        self.pre_calib = pre_calib
        self.l2_normalize = l2_normalize

        # options: robust scaling
        # pre-TS

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor) -> 'Calibrator':
        if self.pre_calib:
            self.pre_calib_: Calibrator = sklearn.base.clone(self.pre_calib)
            self.pre_calib_.fit_torch(y_pred, y_true_labels)
            y_pred = self.pre_calib_.predict_proba_torch(y_pred)
        if self.l2_normalize:
            logits = y_pred.get_logits()
            logits = logits - logits.mean(dim=-1, keepdim=True)
            self.factor_ = 1. / (logits.square().mean().sqrt().item() + 1e-8)
            logits = self.factor_ * logits
            y_pred = CategoricalLogits(logits)

        self.calib_: Calibrator = sklearn.base.clone(self.calibrator)
        self.calib_.fit_torch(y_pred, y_true_labels)

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        if self.pre_calib:
            y_pred = self.pre_calib_.predict_proba_torch(y_pred)
        if self.l2_normalize:
            logits = y_pred.get_logits()
            logits = logits - logits.mean(dim=-1, keepdim=True)
            logits = self.factor_ * logits
            y_pred = CategoricalLogits(logits)

        return self.calib_.predict_proba_torch(y_pred)




class SklearnCalibrator(Calibrator):
    def __init__(self, method: str, cv: str):
        super().__init__()
        self.method = method
        self.cv = cv

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        if X.ndim == 1:
            X = np.stack((1.0-X, X), axis=1)

        n_classes = X.shape[-1]
        est = IdentityEstimator(n_classes)
        est.fit(X, y)

        # tried this workaround for deprecation warnings, but it complains in case of missing classes because it still tries to fit the estimator
        # try:
        #     # cv='prefit' option is deprecated from sklearn 1.6, FrozenEstimator was introduced in sklearn 1.6
        #     from sklearn.frozen import FrozenEstimator
        #     self.calib_ = CalibratedClassifierCV(FrozenEstimator(est), method=self.method, cv=[(np.arange(0), np.arange(X.shape[0]))])
        # except ImportError:
        #     self.calib_ = CalibratedClassifierCV(est, method=self.method, cv=self.cv)

        self.calib_ = CalibratedClassifierCV(est, method=self.method, cv=self.cv)
        self.calib_.fit(X, y)
        self.classes_ = est.classes_

    def predict_proba(self, X):
        if X.ndim == 1:
            X = np.stack((1.0-X, X), axis=1)

        return self.calib_.predict_proba(X)


class WrapCalibratorAsClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, calib: BaseEstimator):
        self.calib = calib

    def fit(self, X, y):
        # print(f'Fitting calibrator: {self.calib}')
        self.calib_ = sklearn.base.clone(self.calib)
        self.calib_.fit(X, y)
        self.classes_ = list(range(np.max(y) + 1))
        return self

    def predict_proba(self, X):
        return self.calib_.predict_proba(X)


def logloss_np(y_true: np.ndarray, y_proba: np.ndarray):
    return -np.mean(np.take_along_axis(np.log(y_proba), y_true[:, None], axis=1))


class DirichletCalibrator(Calibrator):
    def __init__(self, n_cv: int = 0, use_odir: bool = False,
                 reg_lambda: float = 0.0, reg_mu: Optional[float] = None,
                 reg_lambda_grid: Optional[List[float]] = None, reg_mu_grid: Optional[List[float]] = None):
        super().__init__()

        self.n_cv = n_cv
        self.use_odir = use_odir
        self.reg_lambda = reg_lambda
        self.reg_mu = reg_mu
        self.reg_lambda_grid = reg_lambda_grid
        self.reg_mu_grid = reg_mu_grid

    def _fit_impl(self, X: np.ndarray, y: np.ndarray) -> None:
        from dirichletcal.calib.fulldirichlet import FullDirichletCalibrator
        if self.n_cv == 0:
            self.cal_ = FullDirichletCalibrator(reg_lambda=self.reg_lambda, reg_mu=self.reg_mu)
        elif self.n_cv >= 2:
            reg_lambda_grid = self.reg_lambda_grid or [1e-1, 1e-2, 1e-3, 1e-4, 1e-5]
            reg_mu_grid = self.reg_mu_grid or [1e-1, 1e-2, 1e-3, 1e-4, 1e-5]
            calibrator = FullDirichletCalibrator(reg_lambda=reg_lambda_grid,
                                                 reg_mu=reg_mu_grid if self.use_odir else None)
            calibrator = WrapCalibratorAsClassifier(calibrator)
            skf = StratifiedKFold(n_splits=self.n_cv, shuffle=True, random_state=0)
            self.cal_ = GridSearchCV(calibrator, param_grid={'calib__reg_lambda': reg_lambda_grid,
                                                             'calib__reg_mu': reg_mu_grid if self.use_odir else [None]},
                                     cv=skf,
                                     # use this because scikit-learn's logloss scorer fails in case of missing classes
                                     scoring=lambda est, X_test, y_test: -logloss_np(y_test, est.predict_proba(X_test)))
        else:
            raise ValueError(f'n_cv must be either 0 or >=2, but is {self.n_cv}')

        self.cal_.fit(X, y)

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        return self.cal_.predict_proba(X)


class MixtureCalibrator(Calibrator):
    def __init__(self, calibrator: Calibrator, output_constant: float = 1.0, input_constant: float = 0.0):
        super().__init__()
        self.calibrator = calibrator
        self.output_constant = output_constant
        self.input_constant = input_constant

    def _get_mixture(self, dist_1: CategoricalDistribution, unif_coef: float):
        probs = dist_1.get_probs()
        unif_probs = (1.0 / probs.shape[-1]) * torch.ones_like(probs)
        return CategoricalProbs((1.0 - unif_coef) * probs + unif_coef * unif_probs)

    def _fit_torch_impl(self, y_pred: CategoricalDistribution, y_true_labels: torch.Tensor):
        self.n_samples_ = y_pred.get_n_samples()
        if self.input_constant != 0.0:
            y_pred = self._get_mixture(y_pred, self.input_constant / (self.n_samples_ + 1))
        self.cal_ = sklearn.base.clone(self.calibrator)
        self.cal_.fit_torch(y_pred, y_true_labels)

    def predict_proba_torch(self, y_pred: CategoricalDistribution) -> CategoricalDistribution:
        if self.input_constant != 0.0:
            y_pred = self._get_mixture(y_pred, self.input_constant / (self.n_samples_ + 1))
        y_pred = self.cal_.predict_proba_torch(y_pred)
        if self.output_constant != 0.0:
            y_pred = self._get_mixture(y_pred, self.output_constant / (self.n_samples_ + 1))

        return y_pred


def get_calibrator(calibration_method: str, calibrate_with_mixture: bool = False,
                   cal_mixture_output_constant: float = 1.0,
                   cal_mixture_input_constant: float = 0.0, **config) -> Calibrator:
    if calibration_method == 'platt':
        cal = SklearnCalibrator(method='sigmoid', cv='prefit')
    elif calibration_method == 'platt-logits':
        cal = ApplyToLogitsCalibrator(SklearnCalibrator(method='sigmoid', cv='prefit'))
    elif calibration_method == 'isotonic':
        cal = SklearnCalibrator(method='isotonic', cv='prefit')
    elif calibration_method == 'ivap':
        cal = VennAbersCalibrator(use_ovo=config.get('va_use_ovo', False))
    elif calibration_method == 'ivap-ovr':
        cal = MulticlassOneVsRestCalibrator(BinaryVennAbersCalibrator())
    elif calibration_method == 'ivap-ovo':
        cal = MulticlassOneVsOneCalibrator(BinaryVennAbersCalibrator())
    elif calibration_method == 'isotonic-naive-ovr':
        # should be the same as 'isotonic', this is just to check
        cal = MulticlassOneVsRestCalibrator(SklearnCalibrator(method='isotonic', cv='prefit'))
    elif calibration_method == 'cir':
        cal = MulticlassOneVsRestCalibrator(CenteredIsotonicRegressionCalibrator())
    elif calibration_method in ['temp-scaling', 'ts-mix']:
        cal = TemperatureScalingCalibrator(opt=config.get('ts_opt', 'bisection'),
                                           max_bisection_steps=config.get('ts_max_bisection_steps', 30),
                                           lr=config.get('ts_lr', 0.1),
                                           max_iter=config.get('ts_max_iter', 200),
                                           use_inv_temp=config.get('ts_use_inv_temp', True),
                                           inv_temp_init=config.get('ts_inv_temp_init', 1 / 1.5))
    elif calibration_method == 'linear-scaling':
        cal = BinaryLogisticCalibrator(type='linear')
    elif calibration_method == 'affine-scaling':
        cal = BinaryLogisticCalibrator(type='affine')
    elif calibration_method == 'quadratic-scaling':
        cal = BinaryLogisticCalibrator(type='quadratic')
    elif calibration_method == 'svs':
        cal = SVSCalibrator(penalty=config.get('svs_penalty', 'ridge'),
                            rho=config.get('svs_rho', 1.0),
                            tau=config.get('svs_tau', 1.0),
                            lambda_intercept=config.get('svs_lambda_intercept', 1.0),
                            lambda_diagonal=config.get('svs_lambda_diagonal', 1.0),
                            max_iter=config.get('svs_max_iter', 500),
                            tol=config.get('svs_tol', 1e-5),
                            opt=config.get('svs_opt', 'saga'),
                            ts_preprocessing=config.get('svs_ts_preprocessing', True),
                            print_init_info=config.get('svs_print_init_info', True))
    elif calibration_method == 'sms':
        cal = SMSCalibrator(penalty=config.get('sms_penalty', 'ridge'),
                            rho=config.get('sms_rho', 1.0),
                            tau=config.get('sms_tau', 1.0),
                            lambda_intercept=config.get('sms_lambda_intercept', 1.0),
                            lambda_diagonal=config.get('sms_lambda_diagonal', 1.0),
                            lambda_off_diagonal=config.get('sms_lambda_off_diagonal', 1.0),
                            max_iter=config.get('sms_max_iter', 500),
                            tol=config.get('sms_tol', 1e-5),
                            opt=config.get('sms_opt', 'saga'),
                            ts_preprocessing=config.get('sms_ts_preprocessing', True),
                            print_init_info=config.get('sms_print_init_info', True))
    elif calibration_method == 'logistic':
        cal = LogisticCalibrator(binary_type=config.get('logistic_binary_type', 'affine'),
                                 multiclass_type=config.get('logistic_multiclass_type', 'sms'))
    elif calibration_method == 'torchunc-ts':
        cal = TorchUncertaintyTemperatureScalingCalibrator(init_val=config.get('ts_init_val', 1),
                                                           lr=config.get('ts_lr', 0.1),
                                                           max_iter=config.get('ts_max_iter', 100))
    elif calibration_method == 'guo-ts':
        cal = GuoTemperatureScalingCalibrator()
    elif calibration_method == 'torchcal-ts':
        cal = TorchcalTemperatureScalingCalibrator()
    elif calibration_method == 'autogluon-ts':
        cal = AutoGluonTemperatureScalingCalibrator(init_val=config.get('ts_init_val', 1),
                                                    max_iter=config.get('ts_max_iter', 200),
                                                    lr=config.get('ts_lr', 0.1))
    elif calibration_method == 'autogluon-inv-ts':
        cal = AutoGluonInverseTemperatureScalingCalibrator(init_val=config.get('ts_init_val', 1),
                                                           max_iter=config.get('ts_max_iter', 200),
                                                           lr=config.get('ts_lr', 0.1))
    elif calibration_method == 'dircal':
        cal = DirichletCalibrator(n_cv=0, reg_lambda=config.get('dircal_reg_lambda', 0.0),
                                  reg_mu=config.get('dircal_reg_mu', None))
    elif calibration_method == 'dircal-cv':
        cal = DirichletCalibrator(n_cv=config.get('dircal_n_cv', 5), use_odir=config.get('dircal_use_odir', False),
                                  reg_lambda_grid=config.get('dircal_reg_lambda_grid', None),
                                  reg_mu_grid=config.get('dircal_reg_mu_grid', None))
    else:
        raise ValueError(f'Unknown calibration method "{calibration_method}"')

    if calibrate_with_mixture or calibration_method.endswith('-mix'):
        cal = MixtureCalibrator(cal, output_constant=cal_mixture_output_constant,
                                input_constant=cal_mixture_input_constant)

    return cal



from typing import Tuple

import torch


def remove_missing_classes(y_pred: torch.Tensor, y: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Removes missing classes from y_pred and y.
    For example, if y_pred.shape[-1] == 4 but y only contains the values 0 and 2,
    the columns y_pred[..., 1] and y_pred[..., 3] will be removed and the values (0, 2) will be mapped to (0, 1).
    :param y_pred: Predictions of shape (n_samples, n_classes) (should be logits
    because probabilities will not be normalized anymore after removing columns).
    :param y: classes of shape (n_samples,)
    :return: y_pred and y with missing classes removed
    """
    # shapes: y_pred should be n_samples x n_classes, y should be n_samples
    n_classes = y_pred.shape[-1]
    counts = torch.bincount(y, minlength=n_classes)
    is_present = counts > 0
    if torch.all(is_present).item():
        # all classes are present, nothing needs to be removed
        return y_pred, y

    num_present = is_present.sum().item()
    reduced_y_pred = y_pred[..., is_present]
    class_mapping = torch.zeros(n_classes, dtype=torch.long, device=y.device)
    class_mapping[is_present] = torch.arange(num_present, dtype=torch.long, device=y.device)
    reduced_y = class_mapping[y]
    # print(f'{is_present=}, {reduced_y_pred.shape=}, {torch.unique(reduced_y)=}')
    return reduced_y_pred, reduced_y



import math
import warnings
import numpy as np
from numba import njit
from typing import Callable, Tuple
from .utils import _fast_shuffle, _soft_threshold, _mcp_threshold

L_FACTOR = 0.25
ETA_DIVISOR = 3.0

# Proximal operators
@njit(fastmath=True, cache=True)
def _prox_mcp(weight: float, step_size: float, lmbda: float) -> float:
    return _mcp_threshold(weight, lmbda, step_size)

@njit(fastmath=True, cache=True)
def _prox_lasso(weight: float, step_size: float, lmbda: float) -> float:
    return _soft_threshold(weight, step_size * lmbda)

@njit(fastmath=True, cache=True)
def _prox_ridge(weight: float, step_size: float, lmbda: float) -> float:
    return weight / (1.0 + 2.0 * step_size * lmbda)

@njit(fastmath=True, cache=True)
def _prox_none(weight: float, step_size: float, lmbda: float) -> float:
    return weight

# SAGA solver
@njit(fastmath=True, cache=True)
def _saga_binary(
        X: np.ndarray, y: np.ndarray, lambdas: np.ndarray, max_iter: int,
        tol: float, prox_op: Callable[[float, float, float], float],
) -> Tuple[np.ndarray, int]:
    
    n, d = X.shape
    
    w = np.zeros(d)
    previous_w = np.zeros(d)

    samples_seen = 1

    gradient_weights_memory = np.zeros(n)
    avg_gradient = np.zeros(d)
    permutation = np.arange(n, dtype=np.int32)

    L = L_FACTOR * np.max(np.sum(X * X, axis=1))
    step_size = 1.0 / (ETA_DIVISOR * L)

    for epoch in range(1, max_iter + 1):
        _fast_shuffle(permutation)
        previous_w[:] = w[:]

        for i in permutation:
            xi, yi = X[i], y[i]

            # Logistic prediction
            logit = 0.0
            for j in range(d):
                logit += w[j] * xi[j]
            prob = 1.0 / (1.0 + math.exp(-logit))

            # Gradient update
            gradient_weight = prob - yi
            delta_grad = gradient_weight - gradient_weights_memory[i]
            gradient_weights_memory[i] = gradient_weight

            for j in range(d):
                update = delta_grad * xi[j]
                w[j] -= step_size * (update + avg_gradient[j])
                if epoch == 1:
                    avg_gradient[j] += (update - avg_gradient[j]) / samples_seen
                else:
                    avg_gradient[j] += update / n
                w[j] = prox_op(w[j], step_size, lambdas[j])

            if epoch == 1:
                samples_seen += 1
        
        converged = True

        for j in range(d):
            if abs(w[j] - previous_w[j]) > tol * abs(w[j]):
                converged = False
                break
        
        if converged:
            return w, epoch
        
    return w, epoch

# Dictionary mapping
SCALAR_PROX_MAP = {
    "mcp": _prox_mcp,
    "lasso": _prox_lasso,
    "ridge": _prox_ridge,
    None: _prox_none,
}

# Public API functions
def fit_affine_scaling(
        X: np.ndarray, y: np.ndarray, penalty: str, reg_intercept: float, max_iter: int, tol: float
    ) -> np.ndarray:
    """Fit affine scaling."""
    lambdas = np.array([reg_intercept, 0.0])

    try:
        prox_op = SCALAR_PROX_MAP[penalty]
    except KeyError:
        raise ValueError(f"Unknown penalty: {penalty}")

    w, last_iter = _saga_binary(X, y, lambdas, max_iter, tol, prox_op)
    if last_iter == max_iter:
        warnings.warn("SAGA reached max_iter without convergence", RuntimeWarning)
    return w

def fit_quadratic_scaling(
    X: np.ndarray, y: np.ndarray, penalty: str, reg_intercept: float, reg_quadratic: float, max_iter: int, tol: float
    ) -> np.ndarray:
    """Fit quadratic scaling."""
    lambdas = np.array([reg_intercept, 0.0, reg_quadratic])

    try:
        prox_op = SCALAR_PROX_MAP[penalty]
    except KeyError:
        raise ValueError(f"Unknown penalty: {penalty}")

    w, last_iter = _saga_binary(X, y, lambdas, max_iter, tol, prox_op)
    if last_iter == max_iter:
        warnings.warn("SAGA reached max_iter without convergence", RuntimeWarning)
    return w

def warm_up_affine_scaling():
    """Warm up affine scaling functions."""
    X = np.random.randn(10, 2).astype(np.float32)
    y = np.random.randint(0, 2, size=10)
    for penalty in [None, 'mcp', 'lasso', 'ridge']:
        _saga_binary(X, y, np.array([0.1, 0.0]), 10, 1e-4, SCALAR_PROX_MAP[penalty])

def warm_up_quadratic_scaling():
    """Warm up quadratic scaling functions."""
    X = np.random.randn(10, 3).astype(np.float32)
    y = np.random.randint(0, 2, size=10)
    for penalty in [None, 'mcp', 'lasso', 'ridge']:
        _saga_binary(X, y, np.array([0.1, 0.0, 0.1]), 10, 1e-4, SCALAR_PROX_MAP[penalty])



import numpy as np
from numba import njit

MCP_GAMMA = 3.0

@njit(fastmath=True, cache=True)
def _fast_shuffle(arr: np.ndarray) -> None:
    """In-place accelerated array shuffling."""
    for i in range(arr.shape[0] - 1, 0, -1):
        j = np.random.randint(i + 1)
        arr[i], arr[j] = arr[j], arr[i]

@njit(fastmath=True, inline="always", cache=True)
def _soft_threshold(val: float, thresh: float) -> float:
    """Soft-thresholding of val at level thresh."""
    if val > thresh:
        return val - thresh
    elif val < -thresh:
        return val + thresh
    else:
        return 0.0

@njit(fastmath=True, inline="always", cache=True)
def _mcp_threshold(val: float, lmbda: float, step_size: float, gamma: float = MCP_GAMMA) -> float:
    """MCP-thresholding of val with regularization strength lambda and non-convexity parameter gamma."""
    thresh = lmbda * step_size
    abs_val = abs(val)

    if abs_val <= thresh:
        return 0.0

    gamma_lmbda = gamma * lmbda
    if abs_val > gamma_lmbda:
        return val

    div = 1.0 - step_size / gamma
    if val > thresh:
        return (val - thresh) / div
    else: # val < -thresh
        return (val + thresh) / div

@njit(fastmath=True, cache=True)
def _mcp_penalty(val: float, lmbda: float, gamma: float = MCP_GAMMA) -> float:
    """Computes the value of the MCP penalty."""
    abs_val = abs(val)
    if abs_val <= gamma * lmbda:
        return lmbda * abs_val - (abs_val**2) / (2.0 * gamma)
    else:
        return 0.5 * gamma * lmbda**2



import copy
import numpy as np

def join_dicts(*dicts, allow_overlap: bool = True):
    # Attention: arguments do not commute since later dicts can override entries from earlier dicts!
    result = copy.copy(dicts[0])
    for d in dicts[1:]:
        if not allow_overlap and any([key in result for key in d.keys()]):
            raise ValueError(f'Overlapping keys with allow_overlap=False')
        result.update(d)
    return result

def validate_probabilities(X: np.ndarray) -> None:
    """
    Checks if X is a valid 2D array of probability distributions.
    Raises ValueError if the validation fails.
    """
    if X.ndim != 2:
        raise ValueError(f"Input array must be 2D, but got {X.ndim} dimensions.")
    if np.any(X < 0) or np.any(X > 1):
        raise ValueError("All probability values must be within the [0, 1] range.")
    if not np.all(np.isclose(X.sum(axis=1), 1.0)):
        raise ValueError("Each row of the input array must sum to 1 to be a valid probability distribution.")

def binary_probs_to_logits(probs: np.ndarray) -> np.ndarray:
    """Converts binary probabilities to logits using the logit function (inverse sigmoid).
    Clips logits to the log of the tinyest normal float32 to avoid infinite logit values.
    """
    probs = probs.astype(np.float32)
    # 1D array of probabilities
    if probs.ndim == 1:
        logits = probs
    # 2D array of probability pairs [(p0, p1), ...]
    elif probs.ndim == 2 and probs.shape[1] == 2:
        logits = probs[:, 1]
    else:
        raise ValueError(f"Invalid input shape: {probs.shape}."
                         f"1D array or 2D array with shape (n, 2)")
    
    with np.errstate(divide="ignore"):
        logits = np.log(logits) - np.log1p(-logits)
    thresh = np.log(np.finfo(np.float32).tiny)
    return np.clip(logits, a_min=thresh, a_max=-thresh)

def multiclass_probs_to_logits(probs: np.ndarray) -> np.ndarray:
    """Converts multiclass probabilities to logits using the log function.
    Clips logits to the log of the tinyest normal float32 to avoid infinite logit values.
    """
    # probs = np.atleast_2d(probs) # Ensure we have a 2D array
    probs = probs.astype(np.float32)
    thresh = np.log(np.finfo(np.float32).tiny)
    with np.errstate(divide="ignore"):
        logits = np.log(probs)
    logits = np.clip(logits, a_min=thresh, a_max=None)
    return logits



import math
import warnings
import numpy as np
from numba import njit
from typing import Callable, Tuple
from .utils import _fast_shuffle, _soft_threshold, _mcp_threshold, _mcp_penalty

L_FACTOR = 0.25
ETA_DIVISOR = 3.0

# Proximal operators
@njit(fastmath=True, cache=True)
def _prox_mcp(vec: np.ndarray, k: int, step_size: float, lmbda: float) -> None:
    """Applies MCP proximal operator inplace on each component of a k-dimensional parameter vector.
    """
    for i in range(k):
        vec[i] = _mcp_threshold(vec[i], lmbda, step_size)

@njit(fastmath=True, cache=True)
def _prox_lasso(vec: np.ndarray, k: int, step_size: float, lmbda: float) -> None:
    """Applies LASSO proximal operator inplace on each component of a k-dimensional parameter vector.
    """
    thresh = step_size * lmbda
    for i in range(k):
        vec[i] = _soft_threshold(vec[i], thresh)

@njit(fastmath=True, cache=True)
def _prox_ridge(vec: np.ndarray, k: int, step_size: float, lmbda: float) -> None:
    """Applies ridge proximal operator inplace on each component of a k-dimensional parameter vector.
    """
    factor = 1.0 / (1.0 + 2.0 * step_size * lmbda)
    for j in range(k):
        vec[j] *= factor

@njit(fastmath=True, cache=True)
def _prox_none(vec: np.ndarray, k: int, step_size: float, lmbda: float) -> None:
    pass

# Logit computation
@njit(fastmath=True, cache=True)
def _compute_logits(
    xi: np.ndarray, a: float, v: np.ndarray, b: np.ndarray, k: int,
    logits: np.ndarray, exp_logits: np.ndarray
) -> float:
    """
    Computes numerically stable exp(logits) for a single sample.

    This function modifies logits and exp_logits in-place.

    Returns:
        The sum of the exponentiated logits (sum_exp).
    """
    # Part 1: Compute logits and find the maximum
    max_logit = -np.inf
    for j in range(k):
        logit = (a + v[j]) * xi[j] + b[j]
        logits[j] = logit
        if logit > max_logit:
            max_logit = logit

    # Part 2: Compute stable exp(logits) and their sum
    sum_exp = 0.0
    for j in range(k):
        exp_logit = math.exp(logits[j] - max_logit)
        exp_logits[j] = exp_logit
        sum_exp += exp_logit
    
    return sum_exp

# Stopping functions
@njit(fastmath=True, cache=True)
def _compute_primal_dual_ridge(
    X: np.ndarray, y: np.ndarray, a: float, v: np.ndarray, b: np.ndarray,
    reg_intercept: float, reg_diagonal: float
) -> Tuple[float, float]:
    """Computes primal and (approximate) dual objectives for our structured vector scaling convex problem with ridge regularization.
    The dual is only approximate since global scaling parameter a is not regularized. This sometimes leads to dual_obj > primal_obj.
    """
    n, k = X.shape
    
    primal_loss = 0.0
    dual_loss = 0.0
    s_v = np.zeros(k)
    s_b = np.zeros(k)
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, v, b, k, logits, exp_logits
        )
        
        for j in range(k):
            prob = exp_logits[j] / sum_exp
            alpha_ij = -prob
            dual_loss -= prob * math.log(prob)
            if j == yi:
                alpha_ij += 1.0
                primal_loss -= math.log(prob)
            s_v[j] += alpha_ij * xi[j]
            s_b[j] += alpha_ij

    primal_reg = reg_diagonal * np.dot(v, v) + reg_intercept * np.dot(b, b)
    primal_obj = (primal_loss / n) + primal_reg

    dual_reg = 0.0
    n_squared = float(n * n)
    dual_reg -= (1.0 / (4.0 * reg_diagonal * n_squared)) * np.dot(s_v, s_v)
    dual_reg -= (1.0 / (4.0 * reg_intercept * n_squared)) * np.dot(s_b, s_b)

    dual_obj = (dual_loss / n) + dual_reg
    
    return primal_obj, dual_obj

@njit(fastmath=True, cache=True)
def _compute_primal_dual_lasso(
    X: np.ndarray, y: np.ndarray, a: float, v: np.ndarray, b: np.ndarray,
    reg_intercept: float, reg_diagonal: float
) -> Tuple[float, float]:
    """Computes primal and (approximate) dual objectives for our structured vector scaling convex problem with LASSO regularization.
    The dual is only approximate since global scaling parameter a is not regularized. This sometimes leads to dual_obj > primal_obj.
    """
    n, k = X.shape
    
    primal_loss = 0.0
    dual_loss = 0.0
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, v, b, k, logits, exp_logits
        )
        
        for j in range(k):
            prob = exp_logits[j] / sum_exp
            dual_loss -= prob * math.log(prob)
            if j == yi:
                primal_loss -= math.log(prob)

    primal_reg = reg_diagonal * np.sum(np.abs(v)) + reg_intercept * np.sum(np.abs(b))
    primal_obj = (primal_loss / n) + primal_reg

    dual_reg = 0.0
    dual_obj = (dual_loss / n) + dual_reg

    return primal_obj, dual_obj

@njit(fastmath=True, cache=True)
def _compute_primal_dual_none(
    X: np.ndarray, y: np.ndarray, a: float, v: np.ndarray, b: np.ndarray,
    reg_intercept: float, reg_diagonal: float
) -> Tuple[float, float]:
    """Computes primal objective for our structured vector scaling convex problem without regularization.
    Since no regularization is applied, the dual cannot be evaluated. Only primal is used for stopping.
    """
    n, k = X.shape

    primal_loss = 0.0

    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, v, b, k, logits, exp_logits
        )
        
        primal_loss -= math.log(exp_logits[yi] / sum_exp)

    primal_obj = primal_loss / n

    return primal_obj, -np.inf

@njit(fastmath=True, cache=True)
def _compute_primal_dual_mcp(
    X: np.ndarray, y: np.ndarray, a: float, v: np.ndarray, b: np.ndarray,
    reg_intercept: float, reg_diagonal: float
) -> Tuple[float, float]:
    """Computes primal objective for our structured vector scaling convex problem with MCP regularization.
    With MCP penalty, the dual cannot be evaluated. Only primal is used for stopping.
    """
    n, k = X.shape
    
    primal_loss = 0.0
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)

    for i in range(n):
        xi, yi = X[i], y[i]

        sum_exp = _compute_logits(
            xi, a, v, b, k, logits, exp_logits
        )
        
        primal_loss -= math.log(exp_logits[yi] / sum_exp)

    primal_reg = 0.0
    for j in range(k):
        primal_reg += _mcp_penalty(v[j], reg_diagonal)
        primal_reg += _mcp_penalty(b[j], reg_intercept)

    primal_obj = (primal_loss / n) + primal_reg

    return primal_obj, -np.inf

# SAGA solver
@njit(fastmath=True, cache=True)
def _saga_vector_scaling(
    X: np.ndarray, y: np.ndarray, init_scaling: float, reg_intercept: float, reg_diagonal: float,
    max_iter: int, tol: float, prox_op: Callable[[np.ndarray, int, float, float], None], stop_op: Callable[..., float]
) -> Tuple[float, np.ndarray, np.ndarray, int]:
    """Solves our structured vector scaling convex optimization problem using SAGA.

    Args:
        X (np.ndarray): un-calibrated logits.
        y (np.ndarray): un-calibrated labels (not one-hot encoded).
        init_scaling (float): initial value for global scaling parameter.
        reg_intercept (float): regularization strenght applied to intercept parameters.
        reg_diagonal (float): regularization strenght applied to diagonal parameters.
        max_iter (int): max number of dataset passes before SAGA is stopped.
        tol (float): tolerance used for stopping.
            Stopping is based on relative difference between primal and (approximate) dual objectives when available.
            Otherwise relative difference between consecutive primal objectives is used.
        prox_op (Callable[[np.ndarray, int, float, float], None]): proximal operator used to regularize.
        stop_op (Callable[..., float]): function that computes the primal and dual objectives for stopping.

    Returns:
        Tuple[float, np.ndarray, np.ndarray, int]: global scaling parameter a, vector scaling parameter v, intercept parameter b, number of epochs performed.
    """
    n, k = X.shape
    
    a = init_scaling
    v = np.zeros(k)
    b = np.zeros(k)

    previous_objective = np.inf
    samples_seen = 1

    gradient_weights_memory = np.zeros((n, k))
    permutation = np.arange(n, dtype=np.int32)
    avg_gradient_a = 0.0
    avg_gradient_v = np.zeros(k)
    avg_gradient_b = np.zeros(k)
    
    logits = np.zeros(k)
    exp_logits = np.zeros(k)
    delta_grad = np.zeros(k)

    L = L_FACTOR * np.max(np.sum(X * X, axis=1))
    step_size = 1.0 / (ETA_DIVISOR * L)

    for epoch in range(1, max_iter + 1):
        _fast_shuffle(permutation)
        
        for i in permutation:
            xi, yi = X[i], y[i]

            sum_exp = _compute_logits(
                xi, a, v, b, k, logits, exp_logits
            )

            update_a = 0.0
            for j in range(k):
                prob = exp_logits[j] / sum_exp
                gradient_weight = prob - (1.0 if j == yi else 0.0)

                delta_grad[j] = gradient_weight - gradient_weights_memory[i, j]
                gradient_weights_memory[i, j] = gradient_weight
                
                update_a += delta_grad[j] * xi[j]

                update_v = delta_grad[j] * xi[j]
                v[j] -= step_size * (update_v + avg_gradient_v[j])
                if epoch == 1:
                    avg_gradient_v[j] += (update_v - avg_gradient_v[j]) / samples_seen
                else:
                    avg_gradient_v[j] += update_v / n

                update_b = delta_grad[j]
                b[j] -= step_size * (update_b + avg_gradient_b[j])
                if epoch == 1:
                    avg_gradient_b[j] += (update_b - avg_gradient_b[j]) / samples_seen
                else:
                    avg_gradient_b[j] += update_b / n

            a -= step_size * (update_a + avg_gradient_a)
            if epoch == 1:
                avg_gradient_a += (update_a - avg_gradient_a) / samples_seen
                samples_seen += 1
            else:
                avg_gradient_a += update_a / n

            prox_op(v, k, step_size, reg_diagonal)
            prox_op(b, k, step_size, reg_intercept)

        primal_obj, dual_obj = stop_op(X, y, a, v, b, reg_intercept, reg_diagonal)
        if dual_obj == -np.inf:
            gap = abs(primal_obj - previous_objective)
            if gap <= tol * max(1.0, abs(primal_obj)):
                return a, v, b, epoch
            previous_objective = primal_obj
        else:
            # Global scaling parameter a is not regularized, the duality gap is not exact and can be negative.
            # For robustness, we track the absolute gap.
            gap = abs(primal_obj - dual_obj)
            if gap <= tol * max(1.0, abs(primal_obj)):
                return a, v, b, epoch

    return a, v, b, max_iter

# Dictionary mappings
VECTOR_PROX_MAP = {
    "mcp": _prox_mcp,
    "lasso": _prox_lasso,
    "ridge": _prox_ridge,
    None: _prox_none,
}

VECTOR_STOP_MAP = {
    "mcp": _compute_primal_dual_mcp,
    "lasso": _compute_primal_dual_lasso,
    "ridge": _compute_primal_dual_ridge,
    None: _compute_primal_dual_none,
}

# Public API functions
def fit_vector_scaling(
        X: np.ndarray, y: np.ndarray, init_scaling: float, penalty: str, reg_diagonal: float,
        reg_intercept: float, max_iter: int, tol: float
) -> np.ndarray:
    """Public API to fit vector scaling using SAGA."""
    try:
        prox_op = VECTOR_PROX_MAP[penalty]
        stop_op = VECTOR_STOP_MAP[penalty]
    except KeyError:
        raise ValueError(f"Unknown penalty: {penalty}")

    a, v, b, last_iter = _saga_vector_scaling(
        X, y, init_scaling, reg_intercept, reg_diagonal,
        max_iter, tol, prox_op, stop_op
    )
    if last_iter == max_iter:
        warnings.warn("SAGA reached max_iter without convergence", RuntimeWarning)
    
    return a, v, b

def warm_up_vector_scaling():
    """Public API to warm up vector scaling."""
    X = np.random.randn(10, 3).astype(np.float32)
    y = np.random.randint(0, 3, size=10)
    for penalty in [None, 'mcp', 'lasso', 'ridge']:
        _saga_vector_scaling(X, y, 1.0/1.5, 3.0/10.0, 3.0/10.0, 10, 1e-5, VECTOR_PROX_MAP[penalty], VECTOR_STOP_MAP[penalty])



from pathlib import Path
from typing import Union, Optional, List

import numpy as np
import torch


# problem: should a distribution object store the tensor which it does computations on?
# In favor:
# then it knows whether it stores logits or probabilities etc.
# we only need to pass/return one object
# we can have distributions that are mixtures of other distributions
# Counterarguments:
# then batching might be more difficult (but do we need this?) -> could implement an indexing function?
# then we need to store a "constructor function" instead of just storing the constructed object.
# but maybe the "constructor function" can just be the class itself.
# or the heads just need to be implemented so that they know how to construct the distribution

def to_one_hot(y: torch.Tensor, n_classes: int, label_smoothing_eps: float = 0.0) -> torch.Tensor:
    """
    Creates a one-hot representation of y.
    :param y: Labels, shape (n_samples,).
    :param n_classes: Number of classes (needs to be specified because it might be larger than max(y)+1).
    :param label_smoothing_eps: Epsilon parameter for label smoothing.
    :return: Returns a one-hot representation of shape (n_samples, num_classes).
    """
    one_hot = torch.nn.functional.one_hot(y, n_classes).float()
    if label_smoothing_eps > 0.0:
        low = label_smoothing_eps / n_classes
        high = 1.0 - label_smoothing_eps + low
        return low + (high - low) * one_hot
    else:
        return one_hot


class Distribution:
    pass  # abstract base class

    # maybe have get_metrics() and save() function

    def get_metric_names(self) -> List[str]:
        pass  # indicate which metrics are applicable

    def save(self, folder: Union[str, Path]):
        # todo: implement
        pass  # need a way to serialize the distribution (for storing predictions). Typically just name and tensor?

    @staticmethod
    def load(folder: Union[str, Path]) -> 'Distribution':
        # todo: implement
        pass

    def get_n_samples(self) -> int:
        pass

    def __getitem__(self, *args) -> 'Distribution':
        raise NotImplementedError()


class CategoricalDistribution(Distribution):
    # probably need one version that takes logits and one that takes probabilities
    # or even more versions for different link functions?
    def get_logits(self) -> torch.Tensor:
        raise NotImplementedError()

    def get_probs(self) -> torch.Tensor:
        raise NotImplementedError()

    def get_modes(self) -> torch.Tensor:
        raise NotImplementedError()

    def get_n_classes(self) -> int:
        raise NotImplementedError()

    def get_n_samples(self) -> int:
        raise NotImplementedError()

    def is_dirac(self) -> bool:
        raise NotImplementedError()

    def __getitem__(self, *args) -> 'CategoricalDistribution':
        raise NotImplementedError()

    def get_metric_names(self) -> List[str]:
        pass  # todo: implement

    @staticmethod
    def from_labels(labels: torch.Tensor, n_classes: Optional[int] = None,
                    ls_eps: float = 0.0) -> 'CategoricalDistribution':
        if ls_eps == 0.0:
            return CategoricalDirac(labels=labels, n_classes=n_classes)

        if n_classes is None:
            n_classes = torch.max(labels).item() + 1
        return CategoricalProbs(probs=to_one_hot(labels, n_classes=n_classes, label_smoothing_eps=ls_eps))

    @staticmethod
    def from_mixture(distributions: List['CategoricalDistribution'], weights: List[float]) -> 'CategoricalDistribution':
        assert len(distributions) >= 1
        return CategoricalProbs(
            probs=sum([weight * dist.get_probs() for weight, dist in zip(weights, distributions)]))


class CategoricalLogits(CategoricalDistribution):
    # probably need one version that takes logits and one that takes probabilities
    # or even more versions for different link functions?
    def __init__(self, logits: torch.Tensor):
        """
        :param logits: Logits of shape (n_samples, n_classes) or (n_samples,).
            In the latter case, logits are interpreted as binary logits
            such that sigmoid(logits) is the probability for class 1.
            They will be converted to (n_samples, 2) shaped logits.
        """
        if len(logits.shape) == 1:
            logits = torch.stack([torch.zeros_like(logits), logits], dim=1)
        self.logits = logits
        self.probs = None
        self.modes = None

    def get_logits(self) -> torch.Tensor:
        return self.logits

    def get_probs(self) -> torch.Tensor:
        if self.probs is None:
            self.probs = torch.nn.functional.softmax(self.logits, dim=-1)
        return self.probs

    def get_modes(self) -> torch.Tensor:
        # returns a tensor of shape (batch_dims)
        if self.modes is None:
            self.modes = self.logits.argmax(dim=-1)
        return self.modes

    def get_n_classes(self) -> int:
        return self.logits.shape[-1]

    def get_n_samples(self) -> int:
        return self.logits.shape[-2]

    def is_dirac(self) -> bool:
        return False

    def __getitem__(self, *args) -> 'CategoricalDistribution':
        return CategoricalLogits(self.logits.__getitem__(*args))


class CategoricalProbs(CategoricalDistribution):
    # probably need one version that takes logits and one that takes probabilities
    # or even more versions for different link functions?
    def __init__(self, probs: torch.Tensor):
        """
        :param probs: Probabilities of shape (n_samples, n_classes) or (n_samples,).
            In the latter case, the vector is interpreted as probabilities for class 1.
            It will then be converted to a (n_samples, 2) shaped vector.
        """
        if len(probs.shape) == 1:
            probs = torch.stack([1.-probs, probs], dim=1)
        self.logits = None
        self.probs = probs
        self.modes = None
        self.thresh = np.log(np.finfo(np.float32).tiny)

    def get_logits(self) -> torch.Tensor:
        if self.logits is None:
            self.logits = torch.clip(torch.log(self.probs), min=self.thresh)
        return self.logits

    def get_probs(self) -> torch.Tensor:
        return self.probs

    def get_modes(self) -> torch.Tensor:
        # returns a tensor of shape (batch_dims)
        if self.modes is None:
            self.modes = self.probs.argmax(dim=-1)
        return self.modes

    def get_n_classes(self) -> int:
        return self.probs.shape[-1]

    def get_n_samples(self) -> int:
        return self.probs.shape[-2]

    def is_dirac(self) -> bool:
        return False

    def __getitem__(self, *args) -> 'CategoricalDistribution':
        return CategoricalProbs(self.probs.__getitem__(*args))


class CategoricalDirac(CategoricalDistribution):
    def __init__(self, labels: torch.Tensor, n_classes: Optional[int]):
        """
        labels should have shape (n_samples,)
        :param labels:
        :param n_classes:
        """
        assert len(labels.shape) == 1
        if n_classes is None:
            n_classes = torch.max(labels).item() + 1
        self.n_classes = n_classes
        self.labels = labels
        self.probs = None
        self.logits = None
        self.thresh = np.log(np.finfo(np.float32).tiny)

    def get_logits(self) -> torch.Tensor:
        if self.logits is None:
            self.probs = to_one_hot(self.labels, self.n_classes)
            self.logits = torch.clip(torch.log(self.probs), min=self.thresh)
        return self.logits

    def get_probs(self) -> torch.Tensor:
        if self.probs is None:
            self.probs = to_one_hot(self.labels, self.n_classes)
        return self.probs

    def get_modes(self) -> torch.Tensor:
        return self.labels

    def is_dirac(self) -> bool:
        return True

    def get_n_classes(self) -> int:
        return self.n_classes

    def get_n_samples(self) -> int:
        return self.labels.shape[0]

    def __getitem__(self, *args) -> 'CategoricalDirac':
        return CategoricalDirac(labels=self.labels.__getitem__(*args), n_classes=self.get_n_classes())


class ContinuousDistribution:
    def get_quantiles(self, quantiles: torch.Tensor):
        pass

    def get_expectiles(self):
        pass

    def get_mean(self):
        pass

    def pdf(self, x):
        pass

    def cdf(self, x):
        pass

    def int_cdf(self, x):
        pass

    def int_sq_cdf_up_to(self, x):
        pass

    def int_sq_cdf_from(self, x):
        pass

    # CDF, PDF, log-likelihood, integral over (squared) CDF / 1-CDF
    # maybe a way to indicate which quantities are available?
    # maybe a way to indicate which metrics should be evaluated?

    def is_continuous(self):
        pass  # or similar
